// SPDX-License-Identifier: GPL-2.0
/*
 * Sony ifrtp SFP Ethernet device driver
 *
 * Copyright (c) 2023 Sony Corporation. All rights reserved.
 */
#include "ifrtp_logging.h"
#include "i2c.h"
#include <linux/of.h>
#include <linux/fs.h>
#include <linux/i2c.h>
#include <linux/nvmem-consumer.h>

static struct i2c_client *pca9555_client;

static inline s32 _ifrtp_i2c_pca9555_setdirection(struct i2c_client *client, int module_id, unsigned char direction)
{
	return i2c_smbus_write_byte_data(client, 0x06 + module_id, direction);
}

static inline s32 _ifrtp_i2c_pca9555_read(struct i2c_client *client, int module_id)
{
	return i2c_smbus_read_byte_data(client, module_id);
}

static inline s32 _ifrtp_i2c_pca9555_write(struct i2c_client *client, int module_id, unsigned char data)
{
	return i2c_smbus_write_byte_data(client, 0x02 + module_id, data);
}

int ifrtp_i2c_init_pca9555(void)
{
	int i, res;

	struct i2c_adapter *adapter = i2c_get_adapter(PCA9555_BUS);

	if (!adapter) {
		IFRTP_LOG_ERR("Error get I2C adapter %d\n", PCA9555_BUS);
		return -ENODEV;
	}

	pca9555_client = i2c_new_dummy_device(adapter, PCA9555_ADR);
	if (IS_ERR(pca9555_client)) {
		IFRTP_LOG_ERR("Error creating I2C dummy device 0x%02x\n", PCA9555_ADR);
		return PTR_ERR(pca9555_client);
	}

	for (i = 0; i < 2; i++) {
		res = _ifrtp_i2c_pca9555_setdirection(pca9555_client, i, 0x25);
		if (res < 0) {
			IFRTP_LOG_ERR("Error setting i2c direction for module %d: %d\n", i, res);
			return res;
		}

		// Disable TX_DIS
		res = _ifrtp_i2c_pca9555_write(pca9555_client, i, 0x0);
		if (res < 0) {
			IFRTP_LOG_ERR("Error writing to module i2c %d: %d\n", i, res);
			return res;
		}
	}

	return 0;
}

void ifrtp_i2c_release_pca9555(void)
{
	if (!IS_ERR(pca9555_client))
		i2c_unregister_device(pca9555_client);
}

s32 ifrtp_i2c_module_read(int module_id)
{
	if (IS_ERR(pca9555_client))
		return PTR_ERR(pca9555_client);

	return _ifrtp_i2c_pca9555_read(pca9555_client, module_id);
}

int ifrtp_i2c_read_mac_addr(int id, unsigned char *dev_addr)
{
	int res;
	struct nvmem_device *at24_nvmem = nvmem_device_find(EEPROM1_NVMEM_NAME, device_match_name);

	if (IS_ERR(at24_nvmem)) {
		IFRTP_LOG_ERR("Failed to find nvmem device\n");
		return PTR_ERR(at24_nvmem);
	}

	res = nvmem_device_read(at24_nvmem, EEPROM1_MACADDR_IGW + id * EEPROM1_MACADDR_SIZE, 6, dev_addr);
	if (res < 0)
		IFRTP_LOG_ERR("Filed to read eeprom: %d\n", res);
	else
		IFRTP_LOG_TRACE("%d bytes read from eeprom\n", res);

	nvmem_device_put(at24_nvmem);
	return res;
}

