// SPDX-License-Identifier: GPL-2.0
/*
 * Sony ifrtp SFP Ethernet device driver
 *
 * Copyright (c) 2023 Sony Corporation. All rights reserved.
 */
#include <linux/module.h>
#include <linux/init.h>
#include <linux/moduleparam.h>

#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/interrupt.h>
#include <linux/timer.h>
#include <linux/delay.h>
#include <linux/workqueue.h>

#include <linux/in.h>
#include <linux/etherdevice.h>
#include <linux/ip.h>
#include <linux/if_arp.h>
#include <linux/tcp.h>
#include <linux/skbuff.h>
#include <linux/version.h>
#include <linux/dma-direction.h>
#include <linux/dma-mapping.h>
#include <linux/dmaengine.h>
#include <linux/poison.h>
#include <linux/string.h>
#include <linux/gfp.h>

#include "ifrtp.h"
#include "ifrtp_address.h"
#include "ifrtp_logging.h"
#include "ifrtp_driver.h"
#include "ifrtp_filter.h"
#include "ifrtp_ioctl.h"
#include "ifrtp_mdio.h"
#include "io.h"
#include "igw_driver.h"
#include "igw_netreg.h"
#include "irq.h"
#include <linux/socket.h>
#include <linux/hrtimer.h>
#include <linux/time.h>
#include <linux/ethtool.h>
#include <linux/inetdevice.h>
#include <linux/notifier.h>
#include <linux/irqreturn.h>
#include <linux/netdevice.h>

#include <linux/in6.h>
#include <asm/checksum.h>
#include <linux/processor.h>


extern struct ifrtp_config ifrtp_cfg;

#define IFRTP_TX_DESCRIPTOR_DEPTH (ifrtp_cfg.tx_buffer)
#define IFRTP_RX_DESCRIPTOR_DEPTH (ifrtp_cfg.rx_buffer)

#define IFRTP_DMA_CHANNEL_0       ""
#define IFRTP_DMA_CHANNEL_1       ""

DEFINE_SPINLOCK(tx_desc_lock);
DEFINE_SPINLOCK(rx_desc_lock);
u32 *desc_addr_tx;
u32 *desc_addr_rx, *packet_addr_rx;
u64 desc_addr_tx_phys, desc_addr_rx_phys, packet_addr_rx_phys;
struct dma_chan *dmachan, *dmachan_memcpy;
struct ifrtp_recv_mem g_recv_buff[RTP_SSRC_MAX];

struct napi_struct ifrtp_napi;

#ifdef IFRTP_ENABLE_DEBUGFS
	struct dentry *ifrtp_debugfs_dir;
	struct debugfs_u32_array debugfs_arrary_1;
#endif
u32 debugfs_counter_1;
u32 debugfs_var_1;

static int ifrtp_release(struct net_device *dev)
{
    // release ports, irq
	netif_stop_queue(dev);
	return 0;
}

static int ifrtp_config(struct net_device *dev, struct ifmap *map)
{
	if (dev->flags & IFF_UP)
		return -EBUSY;

	if (map->base_addr != dev->base_addr) {
		pr_warn("Can't change I/O address\n");
		return -EOPNOTSUPP;
	}

	if (map->irq != dev->irq) {
		pr_warn("Can't change IRQ\n");
		return -EOPNOTSUPP;
	}

	return 0;
}

struct net_device *ifrtp_devs[IFRTP_NUMVFS_MAX];

static int ifrtp_open(struct net_device *dev)
{
	netif_start_queue(dev);

	return 0;
}

irqreturn_t ifrtp_napi_schedule(void)
{
	if (ifrtp_get_net_rx_irq_status() && napi_schedule_prep(&ifrtp_napi)) {
		ifrtp_disable_net_rx_irq();
		__napi_schedule_irqoff(&ifrtp_napi);
		return IRQ_HANDLED;
	}
	return IRQ_NONE;
}


static int ifrtp_send_direct_multi_message(struct ifrtp_priv *priv, unsigned char *data, unsigned int len)
{
	static unsigned int wp;
	unsigned int rp;
	int i, retval = 0, batch_write_count = 0;
	u32 transmitted = 0, vlen = len / sizeof(struct ifrtp_msghdr);
	struct ifrtp_msghdr *msg_vec = (struct ifrtp_msghdr *)data;

	spin_lock(priv->tx_desc_lock);

    // no need for something like
    // wp = ifrtp_ior32(NULL, IFRTP_TX + IFRTP_OFFSET_DESCRIPTOR_WP);
    // use static variable to reduce ioread count
	rp = ifrtp_ior32(NULL, IFRTP_TX + IFRTP_OFFSET_DESCRIPTOR_RP);

	for (i = 0; i < vlen; ++i) {
		if (unlikely(rp == ((wp+1) & (ifrtp_cfg.tx_buffer-1)))) {
			rp = ifrtp_ior32(NULL, IFRTP_TX + IFRTP_OFFSET_DESCRIPTOR_RP);
			if (unlikely(rp == ((wp+1) & (ifrtp_cfg.tx_buffer-1)))) {
				// Buffer full
				transmitted = i;
				retval = -EAGAIN;
				goto trans_err;
			}
		}

		priv->desc_addr_tx[wp*4]   = msg_vec[i].msg_addr >> 32;
		priv->desc_addr_tx[wp*4+1] = msg_vec[i].msg_addr & 0xFFFFFFFF;
		priv->desc_addr_tx[wp*4+2] = msg_vec[i].msg_len;
		priv->desc_addr_tx[wp*4+3] = priv->stream_id;

		wp = (wp+1) & (ifrtp_cfg.tx_buffer-1);
		++batch_write_count;
		if (batch_write_count >= ifrtp_cfg.tx_write_batch || wp == 0) {
			dma_sync_single_for_device(priv->dmachan->device->dev,
									   (dma_addr_t)desc_addr_tx_phys + ((wp == 0 ? ifrtp_cfg.tx_buffer : wp) - batch_write_count) * 16,
									   batch_write_count * 16,
									   DMA_TO_DEVICE);
			ifrtp_iow32(NULL, wp, IFRTP_TX + IFRTP_OFFSET_DESCRIPTOR_WP);
			batch_write_count = 0;
		}
	}

	transmitted = vlen;
trans_err:
	if (batch_write_count) {
		dma_sync_single_for_device(priv->dmachan->device->dev,
								   (dma_addr_t)desc_addr_tx_phys + (wp - batch_write_count) * 16,
								   batch_write_count * 16,
								   DMA_TO_DEVICE);
		ifrtp_iow32(NULL, wp, IFRTP_TX + IFRTP_OFFSET_DESCRIPTOR_WP);
	}
	spin_unlock(priv->tx_desc_lock);

	return retval;
}

static netdev_tx_t ifrtp_tx(struct sk_buff *skb, struct net_device *dev)
{
	struct ifrtp_priv *priv = netdev_priv(dev);

	if (unlikely(ntohs(skb->protocol) != 0)) {
		kfree_skb(skb);
		return NETDEV_TX_OK;
	}

	if (unlikely(ifrtp_send_direct_multi_message(priv, skb->data, skb->len) < 0)) {
		netdev_err_once(dev, "descriptor buffer full\n");
		priv->stats.tx_errors++;
		priv->stats.tx_fifo_errors++;
		// sk_buff must NOT be freed if NETDEV_TX_BUSY is returned
		// TODO (guo): properly control netif queue stop/wake
		// return NETDEV_TX_BUSY;
	}

    // No need to check if skb == NULL, consume_skb will do it.
    // dev_consume_skb_irq(skb);
	consume_skb(skb);
	return NETDEV_TX_OK;
}


static void ifrtp_tx_timeout(struct net_device *dev, unsigned int txqueue)
{
	struct ifrtp_priv *priv = netdev_priv(dev);
	struct netdev_queue *txq = netdev_get_tx_queue(dev, 0);

	ifrtp_pr_debug("Transmit timeout at %ld, latency %ld\n", jiffies,
			jiffies - txq->trans_start);
    // TODO (guo) control transmission interrupt?
	priv->stats.tx_errors++;

	netif_wake_queue(dev);
}


static struct net_device_stats *ifrtp_stats(struct net_device *dev)
{
	struct ifrtp_priv *priv = netdev_priv(dev);

	return &priv->stats;
}


// TODO (guo) check if necessery
static int ifrtp_change_mtu(struct net_device *dev, int new_mtu)
{
    // struct ifrtp_priv *priv = netdev_priv(dev);

	if ((new_mtu < dev->min_mtu) || (new_mtu > dev->max_mtu))
		return -EINVAL;

	dev->mtu = new_mtu;
	return 0;
}


static int ifrtp_siocdevprivate(struct net_device *dev, struct ifreq *ifr, void __user *data, int cmd)
{
	return ifrtp_ioctl(dev, ifr, cmd);
}

static const struct net_device_ops ifrtp_netdev_ops = {
	.ndo_open            = ifrtp_open,
	.ndo_stop            = ifrtp_release,
	.ndo_start_xmit      = ifrtp_tx,
#if KERNEL_VERSION(5, 15, 0) <= LINUX_VERSION_CODE
	.ndo_siocdevprivate  = ifrtp_siocdevprivate,
#else
	.ndo_do_ioctl        = ifrtp_ioctl,
#endif
	.ndo_set_config      = ifrtp_config,
	.ndo_get_stats       = ifrtp_stats,
	.ndo_change_mtu      = ifrtp_change_mtu,
	.ndo_tx_timeout      = ifrtp_tx_timeout,
};

void __iomem *rtp_recv_buffer_remaped, *rtp_recv_buffer_remap_offset;


static int ifrtp_poll(struct napi_struct *napi, int budget)
{
	int npackets = 0;

    // Regarding the order of reading rp and wp:
    // this is safe since only one thread is polling
    // read wp later to process more packet
	static unsigned int rp;
    // unsigned int rp = ifrtp_ior32(NULL, IFRTP_RX + IFRTP_OFFSET_DESCRIPTOR_RP);
	unsigned int wp = ifrtp_ior32(NULL, IFRTP_RX + IFRTP_OFFSET_DESCRIPTOR_WP);
	unsigned int cleaned_count = 0;

	while (wp != rp && npackets < budget) {
		dma_sync_single_for_cpu(dmachan->device->dev,
				(dma_addr_t)desc_addr_rx_phys,
				IFRTP_RX_DESCRIPTOR_DEPTH * 16,
				DMA_FROM_DEVICE);

		// TODO (guo): reduce sync size?
		if (wp > rp) {
			dma_sync_single_for_cpu(dmachan->device->dev,
					(dma_addr_t)packet_addr_rx_phys + rp * 2048,
					(wp - rp) * 2048,
					DMA_FROM_DEVICE);
		} else {
			dma_sync_single_for_cpu(dmachan->device->dev,
					(dma_addr_t)packet_addr_rx_phys,
					wp * 2048,
					DMA_FROM_DEVICE);
			dma_sync_single_for_cpu(dmachan->device->dev,
					(dma_addr_t)packet_addr_rx_phys + rp * 2048,
					(IFRTP_RX_DESCRIPTOR_DEPTH - rp) * 2048,
					DMA_FROM_DEVICE);
		}

		while (wp != rp && npackets < budget) {
			struct sk_buff *skb;
			struct ifrtp_priv *priv;
			unsigned int len = desc_addr_rx[4*rp + 2];
			unsigned int type = desc_addr_rx[4*rp + 3] & 0xFF;
			unsigned int media_type = (desc_addr_rx[4*rp + 3] >> 8 & 0XFF);

			if (unlikely(type >= ifrtp_cfg.numvfs)) {
				pr_warn_ratelimited("invalid rx Network Type %d, packet dropped\n", type);
				goto rx_checkout;
			}

			priv = netdev_priv(ifrtp_devs[type*2]);

			if (unlikely(len > RTP_MAX_PACKET_SIZE || len < RTP_MIN_PACKET_SIZE)) {
				pr_warn_ratelimited("invalid packet length %d, packet dropped\n", len);
				priv->stats.rx_length_errors++;
				priv->stats.rx_errors++;
				goto rx_checkout;
			}

			struct ifrtp_msghdr msg = {
				.msg_len  = len,
				.ssrc     = ntohl(*(__be32 *)(packet_addr_rx + 512*rp + 6)),
			};

			if (unlikely(media_type >= RTP_SSRC_MAX || media_type < type * RTP_SSRC_PER_CH)) {
				priv->stats.rx_dropped++;
				ifrtp_pr_debug_ratelimited("potentially wrong ssrc 0x%08x from ch%d, recognized as media type %d\n",
										   msg.ssrc, type, media_type);
				goto rx_checkout;
			}
			msg.msg_addr = priv->recv_buff[media_type].addr;

			skb = __napi_alloc_skb(napi, sizeof(struct ifrtp_msghdr), GFP_ATOMIC | __GFP_NOWARN);
			if (unlikely(!skb)) {
				pr_warn_ratelimited("can't alloc sk_buff, packet dropped\n");
				priv->stats.rx_over_errors++;
				priv->stats.rx_errors++;
				goto rx_checkout;
			}

#ifdef IFRTP_RX_USE_DMAC_MEMCPY
			// perform memcpy with dma channel 7
			// pr_warn_ratelimited("dma_memcpy(0x%08lx, 0x%016llx, %d)\n", msg.msg_addr, packet_addr_rx_phys + 2048*rp + 16, len);
			struct dma_async_tx_descriptor *desc = dmaengine_prep_dma_memcpy(dmachan_memcpy,
												   msg.msg_addr,
												   packet_addr_rx_phys + 2048*rp + 16,
												   len, DMA_CTRL_ACK);
												   // len, DMA_CTRL_ACK | DMA_PREP_INTERRUPT);
			if (!desc) {
				pr_err_ratelimited("Failed to prepare DMA transfer\n");
				priv->stats.rx_errors++;
				goto rx_checkout;
			}

			dma_cookie_t cookie = desc->tx_submit(desc);

			if (dma_submit_error(cookie)) {
				pr_err_ratelimited("Failed to submit DMA transfer\n");
				priv->stats.rx_errors++;
				goto rx_checkout;
			}

			// dma_async_issue_pending(dmachan_memcpy);
			// don't checkout ?

			enum dma_status status = dma_sync_wait(dmachan_memcpy, cookie);

			if (status != DMA_COMPLETE)
				pr_err_ratelimited("Failed to perform DMA transfer\n");
#else
			memcpy(msg.msg_addr + rtp_recv_buffer_remap_offset,
				   (unsigned char *)packet_addr_rx + 2048*rp + 16, len);
#endif

			memcpy(skb_put(skb, sizeof(struct ifrtp_msghdr)), &msg, sizeof(struct ifrtp_msghdr));

			priv->recv_buff[media_type].addr += RTP_MAX_PACKET_SIZE;
			if (priv->recv_buff[media_type].addr >= priv->recv_buff[media_type].end)
				priv->recv_buff[media_type].addr = priv->recv_buff[media_type].begin;

			skb->dev = ifrtp_devs[type*2];
			skb->protocol = 0;
			skb->ip_summed = CHECKSUM_UNNECESSARY; // chucksum is done by FPGA

			// no need to check return value, netif_receive_skb will always success
			netif_receive_skb(skb);

			// no need for something like
			// priv->stats.rx_packets++;
			// priv->stats.rx_bytes += len;
			// FPGA will do the job

rx_checkout:
			++npackets;
			++cleaned_count;
			rp = (rp+1) & (IFRTP_RX_DESCRIPTOR_DEPTH-1);
			if (cleaned_count >= ifrtp_cfg.rx_write_batch) {
				ifrtp_iow32(NULL, rp, IFRTP_RX + IFRTP_OFFSET_DESCRIPTOR_RP);
				cleaned_count = 0;
			}
		}

		wp = ifrtp_ior32(NULL, IFRTP_RX + IFRTP_OFFSET_DESCRIPTOR_WP);
	}

	if (cleaned_count)
		ifrtp_iow32(NULL, rp, IFRTP_RX + IFRTP_OFFSET_DESCRIPTOR_RP);

    // processed all packets within budget
	if (npackets < budget) {
		if (napi_complete_done(napi, npackets))
			ifrtp_enable_net_rx_irq();
	}

	return npackets;
}

static struct workqueue_struct *arp_check_wq;
static void ifrtp_schedule_arp_check(struct work_struct *w);
static DECLARE_DELAYED_WORK(ifrtp_work_arp_check, ifrtp_schedule_arp_check);
/**
 * check arp cache periodicly
 */
static void ifrtp_schedule_arp_check(struct work_struct *w)
{
	int i;

	for (i = 0; i < ifrtp_cfg.numvfs; ++i) {
		if (!(ifrtp_devs[i]->flags & IFF_UP))
			continue;

		struct ifrtp_priv *priv = netdev_priv(ifrtp_devs[i]);
		struct IocNetTxDestMacAddressConfig mac_config = { .ch_no = priv->stream_id };
		struct IocNetTxDestIpAddressConfig ip_config = { .ch_no = priv->stream_id };
		unsigned char eth_dst[6];

		if (priv->is_main_mac) {
			ifrtp_get_tx_dest_primary_mac_address(NULL, &mac_config);
			ifrtp_get_tx_dest_primary_address(NULL, &ip_config);
		} else {
			ifrtp_get_tx_dest_secondary_mac_address(NULL, &mac_config);
			ifrtp_get_tx_dest_secondary_address(NULL, &ip_config);
		}

		if (arp_query(htonl(ip_config.ip_dst_addr),
					  priv->is_main_mac ? 0 : 1,
					  eth_dst)) {
			continue;
		}

		if (memcmp(eth_dst, mac_config.eth_dst, 6) != 0) {
			memcpy(mac_config.eth_dst, eth_dst, sizeof(eth_dst));
			if (priv->is_main_mac)
				ifrtp_set_tx_dest_primary_mac_address(NULL, &mac_config);
			else
				ifrtp_set_tx_dest_secondary_mac_address(NULL, &mac_config);
		}
	}
	queue_delayed_work(arp_check_wq, &ifrtp_work_arp_check, HZ * 120);
}

static int ifrtp_get_coalesce(struct net_device *netdev,
							  struct ethtool_coalesce *ec,
							  struct kernel_ethtool_coalesce *kernel_coal,
							  struct netlink_ext_ack *extack)
{
	return -ENOTSUPP;
}

static int ifrtp_set_coalesce(struct net_device *netdev,
							  struct ethtool_coalesce *ec,
							  struct kernel_ethtool_coalesce *kernel_coal,
							  struct netlink_ext_ack *extack)
{
	return -ENOTSUPP;
}

static const struct ethtool_ops ifrtp_ethtool_ops = {
	.supported_coalesce_params = ETHTOOL_COALESCE_RX_USECS,
	.get_coalesce              = ifrtp_get_coalesce,
	.set_coalesce              = ifrtp_set_coalesce,
};

unsigned int ifrtp_read_package_counter(unsigned int counter_top, struct ifrtp_packet_count *pktcnt)
{
	unsigned int exist;

	exist = ifrtp_ior32(NULL, counter_top + IFRTP_OFFSET_PKT_COUNTER_EXIST) & IFRTP_PKT_COUNTER_EXIST_MASK;
	if (exist) {
		ifrtp_memcpy_fromio(NULL, (unsigned int *)pktcnt,
							  counter_top + IFRTP_OFFSET_PKT_COUNTER_PKTCNT,
							  sizeof(struct ifrtp_packet_count) / 4);
	}
	return exist;
}

void ifrtp_update_firewall_settings(void)
{
	static struct xt_table_info *last_private;
	struct xt_table *table;
	struct net_device *dev;
	struct net *net = NULL;

	dev = ifrtp_cfg.driver_handle->netdev[0];
	if (!dev || !(dev->flags & IFF_UP))
		return;

	net = dev_net(dev);
	table = xt_find_table_lock(net, AF_INET, "filter");
	if (IS_ERR(table)) {
		pr_err("error finding xt_table: %ld\n", PTR_ERR(table));
		return;
	}

    // append, delete, insert, etc. from/to iptables will change table->private
    // i.e. replace the old table with new one
	if (table->private != last_private) {
		ifrtp_pr_debug("New xt_table found. Add following rules to Identification module.\n");
		ifrtp_filter_apply_netfilter_udp_port(table->private);
		last_private = table->private;
	}

	xt_table_unlock(table);
}

void ifrtp_update_package_count(void)
{
	struct ifrtp_packet_count pktcnt;
	struct ifrtp_priv *priv;
	int i, j, exist;

	exist = ifrtp_read_package_counter(IFRTP_TX_PKT_COUNTER, &pktcnt);
	if (exist) {
		for (j = 0; j < ifrtp_cfg.numvfs / 2; ++j) {
			if (exist & (1 << j)) {
				priv = netdev_priv(ifrtp_devs[j*2]);
				priv->stats.tx_packets += pktcnt.packet_count[j];
				priv->stats.tx_bytes   += pktcnt.packet_size[j];
			}
		}
	}

	for (i = 0; i < 2; ++i) {
		exist = ifrtp_read_package_counter(IFRTP_RX_PKT_COUNTER + IFRTP_RX_PKT_COUNTER_SUBOFFSET * i, &pktcnt);
		if (exist) {
			for (j = 0; j < ifrtp_cfg.numvfs / 2; ++j) {
				if (exist & (1 << j)) {
					priv = netdev_priv(ifrtp_devs[j*2+i]);
					priv->stats.rx_packets += pktcnt.packet_count[j];
					priv->stats.rx_bytes   += pktcnt.packet_size[j];
				}
			}
		}
	}
}

static struct workqueue_struct *ifrtp_2hz_periodic_wq;
static void ifrtp_2hz_periodic_work(struct work_struct *w);
static DECLARE_DELAYED_WORK(ifrtp_work_2hz_periodic, ifrtp_2hz_periodic_work);
static void ifrtp_2hz_periodic_work(struct work_struct *w)
{
	int i, res;
	static bool prev_rx_los[2] = {};

	for (i = 0; i < 2; ++i) {
		res = ifrtp_i2c_module_read(i);
		if (res < 0) {
			pr_err("error reading sfp module %d: %d", i, res);
		} else if (!(res & 0x20) && prev_rx_los[i]) {
			// RX_LOS 1 -> 0
			if (ifrtp_mdio_link_reset(i, 1000, 2))
				pr_err("error bring up link %d", i);
		}
		prev_rx_los[i] = res & 0x20;
	}

	ifrtp_update_package_count();
	ifrtp_update_firewall_settings();

	queue_delayed_work(ifrtp_2hz_periodic_wq, &ifrtp_work_2hz_periodic, HZ / 2);
}

void ifrtp_init(struct net_device *dev)
{
	struct ifrtp_priv *priv;

	dev->type            = ARPHRD_NONE;
	dev->hard_header_len = 0;
	dev->min_header_len  = 0;
	dev->addr_len        = 0;
	dev->flags           = IFF_POINTOPOINT | IFF_NOARP;
	dev->mtu             = ETH_MAX_MTU;
	dev->min_mtu         = ETH_MIN_MTU;
	dev->max_mtu         = ETH_MAX_MTU;
	dev->tx_queue_len    = 1000; // DEFAULT_TX_QUEUE_LEN
	dev->watchdog_timeo  = 5;
	dev->netdev_ops      = &ifrtp_netdev_ops;
	dev->ethtool_ops     = &ifrtp_ethtool_ops;
	dev->features       |= NETIF_F_HW_CSUM;

	priv = netdev_priv(dev);
	memset(priv, 0, sizeof(struct ifrtp_priv));
	priv->dev = dev;
}


int ifrtp_notifier_call_inetaddr(struct notifier_block *nb, unsigned long event, void *data)
{
	struct in_ifaddr *ifaddr = data;
	struct IocNetTxSourceIpAddressConfig ip_config = {
		.ch_no = 0,
		.ip_dst_addr = 0,
	};

	if (event == NETDEV_UP || event == NETDEV_DOWN) {
		if (event == NETDEV_UP)
			ip_config.ip_dst_addr = ifaddr->ifa_address;

		if (memcmp(ifaddr->ifa_dev->dev->name, IGE_IFNAME"0", sizeof(IGE_IFNAME)) == 0)
			ifrtp_set_tx_source_primary_address(NULL, &ip_config);
		else if (memcmp(ifaddr->ifa_dev->dev->name, IGE_IFNAME"1", sizeof(IGE_IFNAME)) == 0)
			ifrtp_set_tx_source_secondary_address(NULL, &ip_config);
		else
			return 0;
	}

	if (event == NETDEV_UP) {
		// update target mac addr on ip addr change
		queue_delayed_work(arp_check_wq, &ifrtp_work_arp_check, HZ * 4);
	}
	return 0;
}

static struct notifier_block nb = {
	.notifier_call = ifrtp_notifier_call_inetaddr,
};

void ifrtp_cleanup_driver(void)
{
	int i;

	pr_info("cleaning up ifrtp...\n");

	if (arp_check_wq) {
		cancel_delayed_work_sync(&ifrtp_work_arp_check);
		flush_workqueue(arp_check_wq);
		destroy_workqueue(arp_check_wq);
	}

	if (ifrtp_2hz_periodic_wq) {
		cancel_delayed_work_sync(&ifrtp_work_2hz_periodic);
		flush_workqueue(ifrtp_2hz_periodic_wq);
		destroy_workqueue(ifrtp_2hz_periodic_wq);
	}

	napi_disable(&ifrtp_napi);
	for (i = 0; i < IFRTP_NUMVFS_MAX; i++) {
		if (ifrtp_devs[i]) {
			unregister_netdev(ifrtp_devs[i]);
			free_netdev(ifrtp_devs[i]);
		}
	}

	if (dmachan)
		dma_release_channel(dmachan);
	unregister_inetaddr_notifier(&nb);

#ifndef IFRTP_RX_USE_DMAC_MEMCPY
	iounmap(rtp_recv_buffer_remaped);
#else
	if (dmachan_memcpy)
		dma_release_channel(dmachan_memcpy);
#endif

#ifdef IFRTP_ENABLE_DEBUGFS
	debugfs_remove_recursive(ifrtp_debugfs_dir);
#endif
}

static bool ifrtp_dma_channel_match_name(struct dma_chan *chan, void *param)
{
	if (!param || ((char *)param)[0] == '\0')
		return true;
	return strcmp(dma_chan_name(chan), param) == 0;
}

void ifrtp_init_rtp_recv_buffer(struct ifrtp_priv *priv)
{
	int i, j;
	int media_type = priv->stream_id * RTP_SSRC_PER_CH;

	ifrtp_dev_dbg(&priv->dev->dev, "media_type = %d\n", media_type);
	for (i = 0; i < RTP_SSRC_PER_CH; ++i) {
		int media = i + priv->stream_mode * IFRTP_STREAM_BUFFER_MEMAP_NUM_STREAM_TYPE;

		for (j = 0; j < priv->ssrc_count[i]; ++j) {
			priv->recv_buff[media_type].begin = priv->recv_buff[media_type].begin_offset +
				ifrtp_cfg.rtp_recv_buffer[media].start_addr +
				ifrtp_cfg.rtp_recv_buffer[media].ch_offset * (priv->stream_id + j);
			priv->recv_buff[media_type].end = priv->recv_buff[media_type].begin +
				ifrtp_cfg.rtp_recv_buffer[media].size - RTP_MAX_PACKET_SIZE;
			priv->recv_buff[media_type].addr = priv->recv_buff[media_type].begin;

			ifrtp_dev_dbg(&priv->dev->dev, "i = %d, j = %d\n", i, j);
			ifrtp_dev_dbg(&priv->dev->dev, "recv_buff[%d].begin = 0x%08lx\n", media_type, priv->recv_buff[media_type].begin);
			ifrtp_dev_dbg(&priv->dev->dev, "recv_buff[%d].end   = 0x%08lx\n", media_type, priv->recv_buff[media_type].end);
			++media_type;
		}
	}
}

#ifdef IFRTP_ENABLE_DEBUGFS
void ifrtp_init_debugfs(void)
{
	ifrtp_debugfs_dir = debugfs_create_dir("ifrtp", NULL);

	debugfs_arrary_1.n_elements = 2048;
	debugfs_arrary_1.array = kmalloc_array(debugfs_arrary_1.n_elements, sizeof(u32), GFP_KERNEL);
	debugfs_create_u32_array("ifrtp_debugfs_arrary_1", 0444, ifrtp_debugfs_dir, &debugfs_arrary_1);

	debugfs_create_u32("ifrtp_debugfs_counter_1", 0664, ifrtp_debugfs_dir, &debugfs_counter_1);
	debugfs_create_u32("ifrtp_timeout_ioctl_counter", 0664, ifrtp_debugfs_dir, &ifrtp_cfg.timeout_ioctl_debugfs_counter);

	debugfs_create_x32("ifrtp_debugfs_var_1", 0664, ifrtp_debugfs_dir, &debugfs_var_1);
}
#endif

int ifrtp_init_driver(int numvfs)
{
	int i, j, ret = -ENOMEM;

#ifdef IFRTP_ENABLE_DEBUGFS
	ifrtp_init_debugfs();
#endif

	for (i = 0; i < numvfs; ++i) {
		ifrtp_devs[i] = alloc_netdev(sizeof(struct ifrtp_priv), "ivfrtp%d",
				NET_NAME_UNKNOWN, ifrtp_init);
		if (ifrtp_devs[i] == NULL)
			goto error_out;
	}

#if KERNEL_VERSION(6, 2, 0) <= LINUX_VERSION_CODE
	netdev_sw_irq_coalesce_default_on(ifrtp_devs[0]);
#else
	ifrtp_devs[0]->gro_flush_timeout = 200000;
	ifrtp_devs[0]->napi_defer_hard_irqs = 2;
#endif

    // separate napi processing from ksoftirqd
	dev_set_threaded(ifrtp_devs[0], true);

	for (i = 0; i < numvfs;  i++) {
		ret = register_netdev(ifrtp_devs[i]);
		if (ret) {
			pr_err("error %i registering device %s\n",
						  ret, ifrtp_devs[i]->name);
			goto error_out;
		}
	}

#if KERNEL_VERSION(6, 1, 0) <= LINUX_VERSION_CODE
	netif_napi_add_weight(ifrtp_devs[0], &ifrtp_napi, ifrtp_poll, ifrtp_cfg.rtp_poll_weight);
#else
	netif_napi_add(ifrtp_devs[0], &ifrtp_napi, ifrtp_poll, ifrtp_cfg.rtp_poll_weight);
#endif

	dma_cap_mask_t mask;

	dma_cap_zero(mask);
	dma_cap_set(DMA_MEMCPY, mask);
    // TODO (guo): only use 1 dma channel for memcpy for now
	dmachan = dma_request_channel(mask, ifrtp_dma_channel_match_name, IFRTP_DMA_CHANNEL_0);
	if (!dmachan) {
		pr_warn("can't allocate channel %s\n", IFRTP_DMA_CHANNEL_0);
		return -EBUSY;
	}

#ifndef IFRTP_RX_USE_DMAC_MEMCPY
	rtp_recv_buffer_remaped = ioremap_wc(0x64000000, 0x09100000);
	rtp_recv_buffer_remap_offset = rtp_recv_buffer_remaped - 0x64000000;
#else
	dmachan_memcpy = dma_request_channel(mask, ifrtp_dma_channel_match_name, IFRTP_DMA_CHANNEL_1);
	if (!dmachan_memcpy) {
		pr_warn("can't allocate channel %s\n", IFRTP_DMA_CHANNEL_1);
		return -EBUSY;
	}
#endif

    // Allocate descriptor memory
	desc_addr_tx = kmalloc(IFRTP_TX_DESCRIPTOR_DEPTH * 16, GFP_KERNEL | GFP_DMA);
	ifrtp_pr_debug("desc_addr_tx = 0x%px\n", desc_addr_tx);
	if (!desc_addr_tx)
		goto error_out;

	desc_addr_tx_phys = virt_to_phys(desc_addr_tx);
	ifrtp_pr_debug("dma: desc_addr_tx      = 0x%px\n", desc_addr_tx);
	ifrtp_pr_debug("dma: desc_addr_tx_phys = 0x%016llx\n", desc_addr_tx_phys);
	ifrtp_iow32(NULL, IFRTP_TX_DESCRIPTOR_DEPTH, IFRTP_TX + IFRTP_OFFSET_DESCRIPTOR_DEPTH);
	ifrtp_iow32(NULL, desc_addr_tx_phys >> 32, IFRTP_TX + IFRTP_OFFSET_DESCRIPTOR_DAH);
	ifrtp_iow32(NULL, desc_addr_tx_phys & 0xFFFFFFFF, IFRTP_TX + IFRTP_OFFSET_DESCRIPTOR_DAL);
	ifrtp_iow32(NULL, 0, IFRTP_TX + IFRTP_OFFSET_DESCRIPTOR_WP);

	desc_addr_rx = kmalloc(IFRTP_RX_DESCRIPTOR_DEPTH * 16, GFP_KERNEL | GFP_DMA);
	ifrtp_pr_debug("desc_addr_rx = 0x%px\n", desc_addr_rx);
	if (!desc_addr_rx)
		goto error_out;

	desc_addr_rx_phys = virt_to_phys(desc_addr_rx);
	ifrtp_iow32(NULL, IFRTP_RX_DESCRIPTOR_DEPTH, IFRTP_RX + IFRTP_OFFSET_DESCRIPTOR_DEPTH);

	packet_addr_rx = kmalloc(IFRTP_RX_DESCRIPTOR_DEPTH * 2048, GFP_KERNEL | GFP_DMA);
	ifrtp_pr_debug("packet_addr_rx = 0x%px\n", packet_addr_rx);
	if (!packet_addr_rx)
		goto error_out;

	packet_addr_rx_phys = virt_to_phys(packet_addr_rx);
	for (i = 0; i < IFRTP_RX_DESCRIPTOR_DEPTH; ++i) {
		desc_addr_rx[i*4]   = (packet_addr_rx_phys + i*2048) >> 32;
		desc_addr_rx[i*4+1] = (packet_addr_rx_phys + i*2048) & 0xFFFFFFFF;
	}
	dma_sync_single_for_device(dmachan->device->dev,
							   (dma_addr_t)desc_addr_rx_phys,
							   IFRTP_RX_DESCRIPTOR_DEPTH * 16,
							   DMA_TO_DEVICE);

	ifrtp_iow32(NULL, desc_addr_rx_phys >> 32, IFRTP_RX + IFRTP_OFFSET_DESCRIPTOR_DAH);
	ifrtp_iow32(NULL, desc_addr_rx_phys & 0xFFFFFFFF, IFRTP_RX + IFRTP_OFFSET_DESCRIPTOR_DAL);
	ifrtp_iow32(NULL, 0, IFRTP_RX + IFRTP_OFFSET_DESCRIPTOR_RP);

	for (i = 0; i < numvfs;  i++) {
		struct ifrtp_priv *priv = netdev_priv(ifrtp_devs[i]);

		priv->tx_desc_lock = &tx_desc_lock;
		priv->dmachan      = dmachan;
		priv->desc_addr_tx = desc_addr_tx;
		priv->desc_addr_rx = desc_addr_rx;
		priv->stream_id    = i / 2;
		priv->is_main_mac  = !(i & 1);
		priv->stream_mode  = 0;
		priv->recv_buff    = g_recv_buff;
		for (j = 0; j < RTP_SSRC_PER_CH; ++j)
			priv->ssrc_count[j] = 1;

		ifrtp_init_rtp_recv_buffer(priv);
	}

	ifrtp_enable_net_rx_irq();

	napi_enable(&ifrtp_napi);

    // enable force tick for packet counter
	ifrtp_iow32(NULL, 1, IFRTP_TX_PKT_COUNTER + IFRTP_OFFSET_PKT_COUNTER_FTICK);
	ifrtp_iow32(NULL, 1, IFRTP_TX_PKT_COUNTER + IFRTP_OFFSET_PKT_COUNTER_TICK);
	ifrtp_iow32(NULL, 1, IFRTP_RX_PKT_COUNTER + IFRTP_OFFSET_PKT_COUNTER_FTICK);
	ifrtp_iow32(NULL, 1, IFRTP_RX_PKT_COUNTER + IFRTP_OFFSET_PKT_COUNTER_TICK);
	ifrtp_iow32(NULL, 1, IFRTP_RX_PKT_COUNTER + IFRTP_RX_PKT_COUNTER_SUBOFFSET + IFRTP_OFFSET_PKT_COUNTER_FTICK);
	ifrtp_iow32(NULL, 1, IFRTP_RX_PKT_COUNTER + IFRTP_RX_PKT_COUNTER_SUBOFFSET + IFRTP_OFFSET_PKT_COUNTER_TICK);

    // setup workqueue for sfp rx_loss and packet counter
	if (!ifrtp_2hz_periodic_wq)
		ifrtp_2hz_periodic_wq = create_singlethread_workqueue("ifrtp_2hz_periodic");
	if (ifrtp_2hz_periodic_wq)
		queue_delayed_work(ifrtp_2hz_periodic_wq, &ifrtp_work_2hz_periodic, HZ * 1);
	else
		pr_err("error creating workqueue.\n");

    // setup tx arp check workqueue
	if (!arp_check_wq)
		arp_check_wq = create_singlethread_workqueue("ifrtp_arp_check");
	if (arp_check_wq)
		queue_delayed_work(arp_check_wq, &ifrtp_work_arp_check, HZ * 10);
	else
		pr_err("error creating workqueue.\n");

	register_inetaddr_notifier(&nb);
error_out:
	if (ret)
		ifrtp_cleanup_driver();
	return ret;
}

// query mac addr for paddr, using interface igw[ifrtp_if_id], and copy to eth_dst
int arp_query(__be32 paddr, int igw_if_id, char *eth_dst)
{
	struct rtable *rt;
	struct neighbour *n;
	struct net_device *dev;
	struct net *ndev;
	struct in_ifaddr *ifa;
	__be32 haddr, saddr;
	int ret = -EHOSTUNREACH;

	if (paddr == 0)
		return 0;

	ifrtp_pr_debug("ARP Querying IP [%pI4]\n", &paddr);

	dev = ifrtp_cfg.driver_handle->netdev[igw_if_id];
	if (!dev || !(dev->flags & IFF_UP) || !netif_carrier_ok(dev)) {
		pr_err("no available interface for arp\n");
		return ret;
	}

	rcu_read_lock();
	ifa = rcu_dereference(dev->ip_ptr->ifa_list);
	if (ifa) {
		saddr = ifa->ifa_address;
	} else {
		pr_warn("no available source ip addr\n");
		saddr = 0;
	}
	rcu_read_unlock();

	ndev = dev_net(dev);
	rt = ip_route_output(ndev, paddr, saddr, RT_TOS(0), dev->ifindex);
	if (IS_ERR(rt)) {
		pr_err("error reading routing table\n");
		return PTR_ERR(rt);
	}

	if (rt->rt_uses_gateway) {
		if (unlikely(rt->rt_gw_family != AF_INET)) {
			pr_err("arp requires IPv4 to work, rt_gw_family: 0x%02x\n", rt->rt_gw_family);
			return ret;
		}
		haddr = rt->rt_gw4; // default gateway ip
		ifrtp_pr_debug("arp using Default Gateway IP [%pI4]\n", &haddr);
	} else {
		ifrtp_pr_debug("arp rt_uses_gateway: %d\n", rt->rt_uses_gateway);
		haddr = paddr;
	}

	ifrtp_pr_debug("%s haddr = %pI4\n", __func__, &haddr);
	n = __ipv4_neigh_lookup(dev, haddr);
	if (!n)
		n = neigh_create(&arp_tbl, &haddr, dev);

	if (!IS_ERR(n)) {
		ret = 0;
		ifrtp_pr_debug("Target MAC [%pM], nud 0x%02x\n", n->ha, n->nud_state);
		if (neigh_event_send(n, NULL)) {
			int timeout = ifrtp_cfg.arp_timeout_ms;
			int sleep_ms = 10;

			ifrtp_pr_debug("neigh event sent\n");
			while ((n->nud_state & NUD_INCOMPLETE) && timeout > 0) {
				msleep_interruptible(sleep_ms);
				ifrtp_pr_debug("sleep %d ms\n", sleep_ms);
				ifrtp_pr_debug("Target MAC [%pM], nud 0x%02x\n", n->ha, n->nud_state);
				timeout -= sleep_ms;
				sleep_ms *= 2;
			}
			if ((timeout <= 0) || !(n->nud_state & NUD_VALID)) {
				pr_warn("timeout in arp querying %pI4 from %s mac\n", &paddr, igw_if_id ? "sub" : "main");
				ret = -EHOSTUNREACH;
			} else {
				ret = 0;
			}
		}
		ifrtp_pr_debug("Target MAC [%pM], nud 0x%02x\n", n->ha, n->nud_state);
		memcpy(eth_dst, n->ha, sizeof(char[6]));
		neigh_release(n);
		return ret;
	}

	pr_err("error lookup ipv4 neighbour\n");
	return PTR_ERR(n);
}

