// SPDX-License-Identifier: GPL-2.0
/*
 * Sony ifrtp SFP Ethernet device driver
 *
 * Copyright (c) 2023 Sony Corporation. All rights reserved.
 */
#include <linux/types.h>

#include <linux/netfilter/xt_tcpudp.h>
#include <linux/netfilter_ipv4.h>
#include <linux/netfilter_ipv4/ip_tables.h>

#include "ifrtp.h"
#include "ifrtp_filter.h"
#include "igw_netreg.h"

extern struct ifrtp_config ifrtp_cfg;

DEFINE_SPINLOCK(rx_filter_lock);

static const char *verdict_to_string(unsigned int v)
{
	switch (v) {
	case NF_DROP:
		return "DROP";
	case NF_ACCEPT:
		return "ACCEPT";
	case NF_STOLEN:
		return "STOLEN";
	case NF_QUEUE:
		return "QUEUE";
	case NF_REPEAT:
		return "REPEAT";
	case NF_STOP:
		return "STOP";
	default:
		return "ERROR";
	}
}

static const char *proto_to_string(u16 p)
{
	switch (p) {
	case SOL_IP:
		return "all";
	case SOL_UDP:
		return "udp";
	case SOL_TCP:
		return "tcp";
	default:
		return "xxx";
	}
}

bool ifrtp_filter_port_accepted(struct igw_priv *priv, int p)
{
	int i;

	for (i = 0; i < IFRTP_FILTER_MAX_FOR_FW_PER_CORE; ++i) {
		if (priv->filter[i].port == p)
			return true;
	}
	return false;
}

void ifrtp_filter_clean_netfilter_udp_port(struct igw_priv *priv)
{
	int i;

	for (i = 0; i < IFRTP_FILTER_MAX_FOR_FW_PER_CORE; ++i) {
		if (priv->filter[i].netfilter) {
			priv->filter[i].port = 0;
			priv->filter[i].netfilter = false;
		}
	}
}

int ifrtp_filter_get_available_port_slot(struct igw_priv *priv, bool ephemeral_only)
{
	int i;

	if (!ephemeral_only && !priv->filter[0].netfilter)
		priv->last_nf_slot = -1;

	if (!ephemeral_only) {
		priv->last_nf_slot++;
		priv->last_nf_slot %= IFRTP_FILTER_MAX_FOR_FW_PER_CORE;
		return priv->last_nf_slot;
	}

	for (i = priv->last_ephemeral_slot + 1; i != priv->last_ephemeral_slot; i = (i+1) % IFRTP_FILTER_MAX_FOR_FW_PER_CORE) {
		if (!priv->filter[i].netfilter)
			break;
	}
	priv->last_ephemeral_slot = i;
	return i;
}

void ifrtp_filter_apply_netfilter_udp_port(struct xt_table_info *private)
{
	const struct ipt_entry *e;
	unsigned long flags;

	ifrtp_pr_debug("target\tprot\topt\tin\tout\tsource\t\tdestination\n");

	spin_lock_irqsave(&rx_filter_lock, flags);

	ifrtp_filter_clean_netfilter_udp_port(ifrtp_cfg.driver_handle->dev[0]);
	ifrtp_filter_clean_netfilter_udp_port(ifrtp_cfg.driver_handle->dev[1]);

	xt_entry_foreach(e, private->entries, private->size) {
		const struct xt_entry_match *m;
		const struct xt_entry_target *t;
		struct igw_priv *priv;
		unsigned int verdict, i;
		int if_id;

		if (e->ip.proto != SOL_UDP)
			continue;

		t = ipt_get_target((struct ipt_entry *)e);
		// TODO (guo): test unconditional rules (next chain)?
		if (strcmp(t->u.user.name, XT_STANDARD_TARGET) != 0)
			continue;

		verdict = ((struct xt_standard_target *)t)->verdict;
		verdict = -verdict - 1;
		if (verdict != NF_ACCEPT)
			continue;

		if (strncmp(e->ip.iniface, IGE_IFNAME, sizeof(IGE_IFNAME)-1) != 0)
			continue;

		// XXX: this will unfotunately also accept rules applied to interface like "igw0foo",
		// but anyway, such interface shouldn't exist on our device
		if_id = e->ip.iniface[sizeof(IGE_IFNAME)-1] - '0';
		if (if_id < 0 || if_id >= IFRTP_PHY_COUNT)
			continue;

		ifrtp_pr_debug("%s\t%s\t%x\t%s\t%s\t%pI4\t\t%pI4\n", verdict_to_string(verdict), proto_to_string(e->ip.proto), e->ip.flags, e->ip.iniface, e->ip.outiface, &e->ip.src, &e->ip.dst);

		priv = ifrtp_cfg.driver_handle->dev[if_id];

		xt_ematch_foreach(m, e) {
			struct xt_udp *udp = (struct xt_udp *)m->data;

			if (udp->dpts[1] - udp->dpts[0] > IFRTP_FILTER_MAX_FOR_FW_PER_CORE)
				continue;

			for (i = udp->dpts[0]; i <= udp->dpts[1]; ++i) {
				struct IocNetRxPortConfig rx_port_config = {
					.module_no = if_id,
					.filter_no = ifrtp_filter_get_available_port_slot(priv, false) + IFRTP_SDI_CH_MAX_PER_CORE,
					.port = i
				};

				ifrtp_pr_debug("accept udp dpt: %d, filter_no: %d\n", i, rx_port_config.filter_no);
				if (ifrtp_set_rx_port(priv, &rx_port_config)) {
					pr_err("error set rx port %d on slot %d\n", i, rx_port_config.filter_no);
					continue;
				} else {
					priv->filter[rx_port_config.filter_no - IFRTP_SDI_CH_MAX_PER_CORE].port      = i;
					priv->filter[rx_port_config.filter_no - IFRTP_SDI_CH_MAX_PER_CORE].netfilter = true;
				}
			}
		}
	}
	spin_unlock_irqrestore(&rx_filter_lock, flags);
}

void ifrtp_filter_accept_ephemeral_port(struct igw_priv *priv, int port)
{
    // XXX: this won't add ephemeral port if it's already opened,
    // thus introduce some consistency issues in conditions like
    // iptable_accept(p) -> ephemeral_accept(p) -> iptable_clear(p)
    // but it shouldn't be a real problem in our use case
	if (ifrtp_filter_port_accepted(priv, port))
		return;

	spin_lock(&rx_filter_lock);
	{
		struct IocNetRxPortConfig rx_port_config = {
			.module_no = priv->netif_priority,
			.filter_no = ifrtp_filter_get_available_port_slot(priv, true) + IFRTP_SDI_CH_MAX_PER_CORE,
			.port = port
		};

		ifrtp_pr_debug("accept udp ephemeral port: %d, filter_no: %d\n", port, rx_port_config.filter_no);
		if (ifrtp_set_rx_port(priv, &rx_port_config))
			pr_err("error set ephemeral port %d on slot %d\n", port, rx_port_config.filter_no);
		else
			priv->filter[rx_port_config.filter_no - IFRTP_SDI_CH_MAX_PER_CORE].port = port;
	}
	spin_unlock(&rx_filter_lock);
}

