// SPDX-License-Identifier: GPL-2.0
/*
 * Sony ifrtp SFP Ethernet device driver
 *
 * Copyright (c) 2023 Sony Corporation. All rights reserved.
 */
#include "ifrtp.h"
#include "ifrtp_address.h"
#include "ifrtp_driver.h"
#include "ifrtp_ioctl.h"
#include "ifrtp_logging.h"
#include "igw_netreg.h"
#include "igw_net_address.h"
#include "io.h"
#include <linux/timekeeping.h>
#include <linux/jiffies.h>
#include <linux/uaccess.h>
#include <linux/dmaengine.h>
#include <linux/socket.h>
#include <linux/spinlock.h>
#include <linux/errno.h>
#include <linux/ioctl.h>
#include <linux/netdevice.h>


extern struct ifrtp_config ifrtp_cfg;

static int ifrtp_ioc_get_if_addr(struct ifrtp_priv *priv, char *ioc_data)
{
	int result;
	struct sockaddr_in *addr = (struct sockaddr_in *)ioc_data;
	struct IocNetTxSourceIpAddressConfig ip_config = {0};

	if (priv->is_main_mac)
		result = ifrtp_get_tx_source_primary_address(NULL, &ip_config);
	else
		result = ifrtp_get_tx_source_secondary_address(NULL, &ip_config);

	if (result != 0)
		return -EIO;

	struct IocNetTxPortConfig port_config = {
		.ch_no = priv->stream_id
	};
	if (priv->is_main_mac)
		result = ifrtp_get_tx_primary_port(NULL, &port_config);
	else
		result = ifrtp_get_tx_secondary_port(NULL, &port_config);

	if (result != 0)
		return -EIO;

	addr->sin_addr.s_addr = ip_config.ip_dst_addr;
	addr->sin_port = htons(port_config.port >> 16);
	addr->sin_family = AF_INET;
	return 0;
}

static int ifrtp_ioc_set_if_addr(struct ifrtp_priv *priv, char *ioc_data)
{
	struct sockaddr_in *addr = (struct sockaddr_in *)ioc_data;

	if (addr->sin_family != AF_INET)
		return -EINVAL;

    // enable/disable ip filter for ch according to s_addr
    // not thread safe, but fix should not be necessary
	unsigned int dst_ip_addr_en_address = OFFSET_NETRX_IDENT_TOP +
										  OFFSET_NETRX_IDENT_CTL_TOP +
										  (priv->is_main_mac ? 0 : OFFSET_NETRX_IDENT_CTL_SIZE) +
										  OFFSET_NETRX_IDENT_CTL_DST_IP_FILTER_EN;
	unsigned int dst_ip_addr_en = ifrtp_ior32(NULL, dst_ip_addr_en_address);

	if (addr->sin_addr.s_addr == 0) {
		// bind to 0.0.0.0
		// bypass ip addr filter
		dst_ip_addr_en |= 1UL << priv->stream_id;
	} else {
		dst_ip_addr_en &= ~(1UL << priv->stream_id);
	}

	if (ifrtp_iow32(NULL, dst_ip_addr_en, dst_ip_addr_en_address))
		return -EIO;

	struct IocNetRxDestIpAddressConfig rx_ip_config = {
		.module_no = priv->is_main_mac ? 0 : 1,
		.filter_no = priv->stream_id,
		.ip_dst_addr = addr->sin_addr.s_addr
	};

	if (ifrtp_set_rx_dest_address(NULL, &rx_ip_config))
		return -EIO;

	struct IocNetRxPortConfig rx_port_config = {
		.module_no = priv->is_main_mac ? 0 : 1,
		.filter_no = priv->stream_id,
		.port = ntohs(addr->sin_port)
	};
    // TODO check if rollback is necessary on error
	if (ifrtp_set_rx_port(NULL, &rx_port_config))
		return -EIO;

	ifrtp_dev_dbg(&priv->dev->dev, "set stream %d %s mac listen address to %pI4:%d\n",
				  priv->stream_id,
				  priv->is_main_mac ? "main" : "sub",
				  &addr->sin_addr,
				  ntohs(addr->sin_port));

	struct IocNetTxPortConfig tx_port_config = {
		.ch_no = priv->stream_id
	};
	if (priv->is_main_mac ? ifrtp_get_tx_primary_port(NULL, &tx_port_config) != 0 :
							ifrtp_get_tx_secondary_port(NULL, &tx_port_config) != 0) {
		return -EIO;
	}

    // not thread safe, but fix should not be necessary
	tx_port_config.port &= 0xFFFF;
	tx_port_config.port += (ntohs(addr->sin_port) & 0xFFFF) << 16;   // set_tx_primary_port use host endian...

	int retval = 0;
    // set sending port
	spin_lock(priv->tx_desc_lock);
	if (priv->is_main_mac ? ifrtp_set_tx_primary_port(NULL, &tx_port_config) != 0 :
						   ifrtp_set_tx_secondary_port(NULL, &tx_port_config) != 0) {
		retval = -EIO;
	}
	spin_unlock(priv->tx_desc_lock);

	return retval;
}

static int ifrtp_ioc_get_if_dest_addr(struct ifrtp_priv *priv, char *ioc_data)
{
	struct sockaddr_in *addr = (struct sockaddr_in *)ioc_data;

	struct IocNetTxPortConfig port_config = {
		.ch_no = priv->stream_id,
		.port = 0
	};
	if (priv->is_main_mac ? ifrtp_get_tx_primary_port(NULL, &port_config) != 0 :
						   ifrtp_get_tx_secondary_port(NULL, &port_config) != 0)
		return -EIO;

	struct IocNetTxDestIpAddressConfig ip_config = {
		.ch_no = priv->stream_id,
		.ip_dst_addr = 0
	};
	if (priv->is_main_mac ? ifrtp_get_tx_dest_primary_address(NULL, &ip_config) != 0 :
						   ifrtp_get_tx_dest_secondary_address(NULL, &ip_config) != 0)
		return -EIO;

	addr->sin_addr.s_addr = htonl(ip_config.ip_dst_addr);  // ifrtp_get_tx_dest_primary_addr use host endian...
	addr->sin_port = htons(port_config.port & 0xFFFF);
	addr->sin_family = AF_INET;
	return 0;
}

static int ifrtp_ioc_set_if_dest_addr(struct ifrtp_priv *priv, char *ioc_data)
{
	struct sockaddr_in *addr = (struct sockaddr_in *)ioc_data;

	if (addr->sin_family != AF_INET)
		return -EINVAL;

	struct IocNetTxPortConfig port_config = {
		.ch_no = priv->stream_id,
		.port = 0
	};
	if (priv->is_main_mac ? ifrtp_get_tx_primary_port(NULL, &port_config) != 0 :
						   ifrtp_get_tx_secondary_port(NULL, &port_config) != 0)
		return -EIO;
    // not thread safe, but fix should not be necessary
	port_config.port &= 0xFFFF0000;
	port_config.port += ntohs(addr->sin_port) & 0xFFFF;   // ifrtp_set_tx_primary_port use host endian...

	struct IocNetTxDestIpAddressConfig ip_config = {
		.ch_no = priv->stream_id,
		.ip_dst_addr = addr->sin_addr.s_addr    // ifrtp_set_tx_dest_primary_addr use network endian...
	};

	struct IocNetTxDestMacAddressConfig mac_config = {
		.ch_no = priv->stream_id
	};
	if (arp_query(addr->sin_addr.s_addr,
				  priv->is_main_mac ? 0 : 1,
				  mac_config.eth_dst) != 0) {
		// no mac address found for dest ip
		dev_warn(&priv->dev->dev, "host %pI4 unreachable\n", &ip_config.ip_dst_addr);
		return -EHOSTUNREACH;
	}

	ifrtp_dev_dbg(&priv->dev->dev, "set stream %d %s mac destination address to [%pM] %pI4:%d\n",
				  priv->stream_id,
				  priv->is_main_mac ? "main" : "sub",
				  mac_config.eth_dst,
				  &ip_config.ip_dst_addr,
				  ntohs(addr->sin_port));

	int retval = 0;

	spin_lock(priv->tx_desc_lock);
	if (priv->is_main_mac) {
		if (ifrtp_set_tx_primary_port(NULL, &port_config) != 0 ||
			ifrtp_set_tx_dest_primary_address(NULL, &ip_config) != 0 ||
			ifrtp_set_tx_dest_primary_mac_address(NULL, &mac_config) != 0) {
			retval = -EIO;
		}
	} else {
		struct IocNetTxDuplicateEnableConfig dup_config = {
			.ch_no  = priv->stream_id,
			.enable = addr->sin_addr.s_addr == 0 ? 0 : 1,
		};

		if (ifrtp_set_tx_secondary_port(NULL, &port_config) != 0 ||
			ifrtp_set_tx_dest_secondary_address(NULL, &ip_config) != 0 ||
			ifrtp_set_tx_dest_secondary_mac_address(NULL, &mac_config) != 0 ||
			ifrtp_set_tx_duplicate_enable(NULL, &dup_config)) {
			retval = -EIO;
		}
		ifrtp_dev_dbg(&priv->dev->dev, "%sable tx duplication on stream %d\n",
					  dup_config.enable ? "en" : "dis",
					  dup_config.ch_no);
	}
	spin_unlock(priv->tx_desc_lock);
	return retval;
}

static int ifrtp_ioc_set_if_memmap_mode(struct ifrtp_priv *priv, char *ioc_data)
{
	unsigned int stream_mode = *(unsigned int *)ioc_data;

	if (stream_mode >= IFRTP_MEMMAP_NUMBER_OF_TYPES) {
		dev_err(&priv->dev->dev, "invalid memmap mode %u\n", stream_mode);
		return -EINVAL;
	}

    // not thread safe but should not be a problem
	priv->stream_mode = stream_mode;

	ifrtp_dev_dbg(&priv->dev->dev, "set stream %d memmap mode to %d\n",
				  priv->stream_id,
				  priv->stream_mode);

	ifrtp_init_rtp_recv_buffer(priv);

	return 0;
}

static int ifrtp_ioc_set_if_ssrc(struct ifrtp_priv *priv, char *ioc_data)
{
    // not thread safe but should not be a problem
    // this may not be the best practice but...
	int i, ssrc_count;
	u32 ssrc_vec[RTP_SSRC_PER_CH * RTP_QFHD_CH_PER_ST];
	struct _ifrtp_ssrc_config *ssrc = (struct _ifrtp_ssrc_config *)ioc_data;

	ssrc_count = ssrc->v_count + ssrc->a_count + ssrc->m_count;
	if (ssrc_count > RTP_SSRC_PER_CH * RTP_QFHD_CH_PER_ST)
		return -EINVAL;

	if (copy_from_user(ssrc_vec, ssrc->ssrc_vec, sizeof(u32) * ssrc_count))
		return -EFAULT;

	priv->ssrc_count[0] = ssrc->v_count;
	priv->ssrc_count[1] = ssrc->a_count;
	priv->ssrc_count[2] = ssrc->m_count;

	ifrtp_dev_dbg(&priv->dev->dev, "setting stream %d %s ssrc\n",
				  priv->stream_id,
				  priv->is_main_mac ? "main mac" : "sub mac");
	for (i = 0; i < ssrc_count; ++i)
		ifrtp_dev_dbg(&priv->dev->dev, "0x%08x\n", ssrc_vec[i]);

	struct IocPayloadExtractionRxSSRCConfig config = {
		.area_no      = priv->is_main_mac ? 0 : 1,
		.dual_en      = 1,
		.nwk_pkt_type = priv->stream_id,
	};
	struct IocNetRxTrmCtrlBucketInitConfig trm_config = {
		.nwk_pkt_type = 0,
	};
	for (i = 0; i < ssrc_count; ++i) {
		// a dirty workaround to reset fec offset for non-video type
		// TODO (guo): clean this up
		if (i >= ssrc->v_count) {
			priv->recv_buff[RTP_SSRC_PER_CH * priv->stream_id + i].begin_offset = 0;
		}

		config.ssrc       = ssrc_vec[i];
		config.media_type = RTP_SSRC_PER_CH * priv->stream_id + i;
		ifrtp_set_rx_extraction_dual_ssrc(NULL, &config);

		trm_config.nwk_pkt_type |= 1 << config.media_type;
	}

	ifrtp_init_rtp_recv_buffer(priv);

	ifrtp_dev_dbg(&priv->dev->dev, "init trm bucket 0x%08x\n", trm_config.nwk_pkt_type);
	ifrtp_set_rx_trm_ctrl_bucket_init(NULL, &trm_config);

	return 0;
}

static int ifrtp_ioc_set_offset_v(struct ifrtp_priv *priv, char *ioc_data)
{

    // not thread safe but should not be a problem
	priv->recv_buff[priv->stream_id * RTP_SSRC_PER_CH].begin_offset = *(unsigned int *)ioc_data;

	ifrtp_dev_dbg(&priv->dev->dev, "set stream %d fec offset to %lu\n",
				  priv->stream_id,
				  priv->recv_buff[priv->stream_id * RTP_SSRC_PER_CH].begin_offset);
	ifrtp_init_rtp_recv_buffer(priv);
	return 0;
}

static int ifrtp_ioc_read_packet_loss_monitor(struct ifrtp_priv *priv, char *ioc_data)
{
	struct IocNetRxSequenceNumberErrorStatusConfig config = {
		.engine_no = priv->is_main_mac ? 0 : 1,
		.sequence_number_error_mask = 7 << priv->stream_id * 3
	};

	if (ifrtp_get_rx_sequence_number_error_status(NULL, &config))
		return -EIO;

	*(unsigned int *)ioc_data = config.sequence_number_error;
	return 0;
}

int ifrtp_ioctl(struct net_device *dev, struct ifreq *ifr, int cmd)
{
	int result = 0;
	struct _ifrtp_ioc_config config_if;
	unsigned int ioc_cmd = 0x0;
	size_t data_size = 0;
	struct ifrtp_priv *priv;

	char ioc_data[256] = {0};

	u64 ts0 = ktime_get_mono_fast_ns();

	if (cmd != SIOCDEVPRIVATE)
		return -EOPNOTSUPP;

	if (ifr->ifr_data == NULL)
		return -EINVAL;

	if (copy_from_user((void *)&config_if, ifr->ifr_data, sizeof(config_if))) {
		dev_err(&dev->dev, "[IOCTL] Failed to copy ifr_data from user\n");
		return -ENOMEM;
	}

	priv = (struct ifrtp_priv *)netdev_priv(dev);

	ioc_cmd = config_if.cmd;
	data_size = _IOC_SIZE(ioc_cmd);
	if (_IOC_DIR(ioc_cmd) & (_IOC_READ | _IOC_WRITE)) {
		if (data_size > sizeof(ioc_data)) {
			dev_err(&dev->dev, "[IOCTL] Too large ioc_data from user cmd %08x, size %zu\n", ioc_cmd, data_size);
			return -EINVAL;
		}
		if (copy_from_user((void *)ioc_data, config_if.data, data_size)) {
			dev_err(&dev->dev, "[IOCTL] Failed to copy ioc_data from user cmd %08x, size %zu\n", ioc_cmd, data_size);
			return -ENOMEM;
		}
	}

	switch (ioc_cmd) {
	case IFRTP_IOCGIFADDR:
		result = ifrtp_ioc_get_if_addr(priv, ioc_data);
		break;
	case IFRTP_IOCSIFADDR:
		result = ifrtp_ioc_set_if_addr(priv, ioc_data);
		break;
	case IFRTP_IOCGIFDSTADDR:
		result = ifrtp_ioc_get_if_dest_addr(priv, ioc_data);
		break;
	case IFRTP_IOCSIFDSTADDR:
		result = ifrtp_ioc_set_if_dest_addr(priv, ioc_data);
		break;
	case IFRTP_IOCSIFMMAPMOD:
		result = ifrtp_ioc_set_if_memmap_mode(priv, ioc_data);
		break;
	case IFRTP_IOCSIFSSRC:
		result = ifrtp_ioc_set_if_ssrc(priv, ioc_data);
		break;
	case IFRTP_IOCSOFFSETV:
		result = ifrtp_ioc_set_offset_v(priv, ioc_data);
		break;
	case IFRTP_IOCGPLSSFLG:
		result = ifrtp_ioc_read_packet_loss_monitor(priv, ioc_data);
		break;
	default:
		dev_err(&dev->dev, "[IOCTL] Unsupported command. %08x received\n", ioc_cmd);
		return -EOPNOTSUPP;
	}
	if (_IOC_DIR(ioc_cmd) & _IOC_READ) {
		if (copy_to_user((void *)config_if.data, (void *)ioc_data, data_size)) {
			dev_err(&dev->dev, "[IOCTL] Failed to copy ioc_data to user cmd %08x, size %zu\n", ioc_cmd, data_size);
			result = -ENOMEM;
		}
	}

	ts0 = ktime_get_mono_fast_ns() - ts0;
	if (ts0 > ifrtp_cfg.timeout_warn_ns_ioctl) {
		ifrtp_cfg.timeout_ioctl_debugfs_counter++;
		ifrtp_dev_dbg(&dev->dev, "ioctl 0x%08x spent %lluus\n", ioc_cmd, ts0/1000);
	}

	return result;

}
