// SPDX-License-Identifier: GPL-2.0
/*
 * Sony ifrtp SFP Ethernet device driver
 *
 * Copyright (c) 2023 Sony Corporation. All rights reserved.
 */
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/etherdevice.h>
#include <linux/proc_fs.h>
#include <linux/byteorder/generic.h>
#include <linux/inet.h>
#include <linux/interrupt.h>
#include <linux/irqreturn.h>
#include <linux/netdevice.h>
#include <linux/printk.h>
#include <linux/of_platform.h>

#include "i2c.h"
#include "io.h"

#include "igw_netreg.h"
#include "igw_driver.h"

#include "ifrtp.h"
#include "ifrtp_logging.h"
#include "ifrtp_driver.h"
#include "ifrtp_mdio.h"

// default parameter
struct ifrtp_config ifrtp_cfg = {
	.numvfs                    = 16,
	.numphys                   = 2,
	.rtp_poll_weight           = NAPI_POLL_WEIGHT,
	.timeout_warn_ns_ioctl     = 2000000,
	.arp_timeout_ms            = 5000,
	.tx_buffer                 = 4096,
	.rx_buffer                 = 2048,
	.tx_write_batch            = 64,
	.rx_write_batch            = 64,
	.main_mac_mask             = 0x50FFFFFF,
	.stream_buffer_packet_size = 1536,
#ifdef DEBUG
	.debug                 = true
#else
	.debug                 = false,
#endif
};

struct ifrtp_global ifrtp_shared = {
	.cfg = &ifrtp_cfg,
};

module_param_named(numvfs, ifrtp_cfg.numvfs, int, 0444);
MODULE_PARM_DESC(numvfs, " Number of virtual interfaces, should be <= 24. default: 24");

module_param_named(rtp_poll_weight, ifrtp_cfg.rtp_poll_weight, int, 0444);
MODULE_PARM_DESC(rtp_poll_weight, " NAPI poll weight for rtp interface. default: 64");

module_param_named(tx_buffer, ifrtp_cfg.tx_buffer, ushort, 0444);
MODULE_PARM_DESC(tx_buffer, " Number of packets TX buffer can hold, should be < 65536. default: 4096");

module_param_named(rx_buffer, ifrtp_cfg.rx_buffer, ushort, 0444);
MODULE_PARM_DESC(rx_buffer, " Number of packets RX buffer can hold, should be < 65536. default: 2048");

module_param_named(tx_write_batch, ifrtp_cfg.tx_write_batch, ushort, (0664));
MODULE_PARM_DESC(tx_write_batch, " Number of packets to send in one tx wp update, should be less that tx_buffer. default: 16");

module_param_named(rx_write_batch, ifrtp_cfg.rx_write_batch, ushort, (0664));
MODULE_PARM_DESC(rx_write_batch, " Number of packets to release in one rx rp update, should be less that rx_buffer. default: 16");

module_param_named(timeout_warn_ns_ioctl, ifrtp_cfg.timeout_warn_ns_ioctl, uint, (0664));
MODULE_PARM_DESC(ioctl_warn_timeout_ns, " Timeout in ns for ioctl to print warnning. default: 1000000");

module_param_named(arp_timeout_ms, ifrtp_cfg.arp_timeout_ms, uint, (0664));
MODULE_PARM_DESC(arp_timeout_ms, " Timeout of arp request. default: 5000");

module_param_named(fhd_ves_addr,    ifrtp_cfg.rtp_recv_buffer[0].start_addr, uint, 0444);
module_param_named(fhd_ves_offset,  ifrtp_cfg.rtp_recv_buffer[0].ch_offset, uint, 0444);
module_param_named(fhd_ves_size,    ifrtp_cfg.rtp_recv_buffer[0].size, uint, 0444);
module_param_named(fhd_aes_addr,    ifrtp_cfg.rtp_recv_buffer[1].start_addr, uint, 0444);
module_param_named(fhd_aes_offset,  ifrtp_cfg.rtp_recv_buffer[1].ch_offset, uint, 0444);
module_param_named(fhd_aes_size,    ifrtp_cfg.rtp_recv_buffer[1].size, uint, 0444);
module_param_named(fhd_mes_addr,    ifrtp_cfg.rtp_recv_buffer[2].start_addr, uint, 0444);
module_param_named(fhd_mes_offset,  ifrtp_cfg.rtp_recv_buffer[2].ch_offset, uint, 0444);
module_param_named(fhd_mes_size,    ifrtp_cfg.rtp_recv_buffer[2].size, uint, 0444);
module_param_named(qfhd_ves_addr,   ifrtp_cfg.rtp_recv_buffer[3].start_addr, uint, 0444);
module_param_named(qfhd_ves_offset, ifrtp_cfg.rtp_recv_buffer[3].ch_offset, uint, 0444);
module_param_named(qfhd_ves_size,   ifrtp_cfg.rtp_recv_buffer[3].size, uint, 0444);
module_param_named(qfhd_aes_addr,   ifrtp_cfg.rtp_recv_buffer[4].start_addr, uint, 0444);
module_param_named(qfhd_aes_offset, ifrtp_cfg.rtp_recv_buffer[4].ch_offset, uint, 0444);
module_param_named(qfhd_aes_size,   ifrtp_cfg.rtp_recv_buffer[4].size, uint, 0444);
module_param_named(qfhd_mes_addr,   ifrtp_cfg.rtp_recv_buffer[5].start_addr, uint, 0444);
module_param_named(qfhd_mes_offset, ifrtp_cfg.rtp_recv_buffer[5].ch_offset, uint, 0444);
module_param_named(qfhd_mes_size,   ifrtp_cfg.rtp_recv_buffer[5].size, uint, 0444);

MODULE_PARM_DESC(fhd_ves_addr,    "FHD Video RTP Recv Buffer start address.");
MODULE_PARM_DESC(fhd_ves_offset,  "FHD Video RTP Recv Buffer offset per channel.");
MODULE_PARM_DESC(fhd_ves_size,    "FHD Video RTP Recv Buffer size per channel.");
MODULE_PARM_DESC(fhd_aes_addr,    "FHD Audio RTP Recv Buffer start address.");
MODULE_PARM_DESC(fhd_aes_offset,  "FHD Audio RTP Recv Buffer offset per channel.");
MODULE_PARM_DESC(fhd_aes_size,    "FHD Audio RTP Recv Buffer size per channel.");
MODULE_PARM_DESC(fhd_mes_addr,    "FHD Metadata RTP Recv Buffer start address.");
MODULE_PARM_DESC(fhd_mes_offset,  "FHD Metadata RTP Recv Buffer offset per channel.");
MODULE_PARM_DESC(fhd_mes_size,    "FHD Metadata RTP Recv Buffer size per channel.");
MODULE_PARM_DESC(qfhd_ves_addr,   "QFHD Video RTP Recv Buffer start address.");
MODULE_PARM_DESC(qfhd_ves_offset, "QFHD Video RTP Recv Buffer offset per channel.");
MODULE_PARM_DESC(qfhd_ves_size,   "QFHD Video RTP Recv Buffer size per channel.");
MODULE_PARM_DESC(qfhd_aes_addr,   "QFHD Audio RTP Recv Buffer start address.");
MODULE_PARM_DESC(qfhd_aes_offset, "QFHD Audio RTP Recv Buffer offset per channel.");
MODULE_PARM_DESC(qfhd_aes_size,   "QFHD Audio RTP Recv Buffer size per channel.");
MODULE_PARM_DESC(qfhd_mes_addr,   "QFHD Metadata RTP Recv Buffer start address.");
MODULE_PARM_DESC(qfhd_mes_offset, "QFHD Metadata RTP Recv Buffer offset per channel.");
MODULE_PARM_DESC(qfhd_mes_size,   "QFHD Metadata RTP Recv Buffer size per channel.");

module_param_named(debug_log, ifrtp_cfg.debug, bool, (0664));
MODULE_PARM_DESC(debug_log, " Enable debug log print. default: false");

static struct igw_driver_handle driver_handle;

static void mark_raw_packet_received(unsigned long val)
{
	bool reschedule = false;
	struct igw_driver_handle *handle = (struct igw_driver_handle *)val;

	if (handle->raw_packet_received) {
		igw_net_raw_packet_isr(handle->dev[0]);
		if (ifrtp_irq_get_raw_packet(handle->dev[0])) {
			reschedule = true;
		} else {
			handle->raw_packet_received = false;
			ifrtp_irq_enable_raw_packet(handle->dev[0]);
		}
	}
	if (reschedule)
		tasklet_schedule(&handle->tasklet_raw_packet);
}


static int interrupt_tasklet_setup(struct igw_driver_handle *handle)
{
	tasklet_init(&handle->tasklet_raw_packet, mark_raw_packet_received, (unsigned long)handle);
	return 0;
}


static int igw_init_driver(void)
{
	int i, core_idx, dev_idx;
	struct igw_priv *priv = NULL;
	int err;

	driver_handle.netif_index_max = 2;

	for (i = 0; i < driver_handle.netif_index_max; ++i) {
		char drv_name[sizeof(IFRTP_IFNAME) + 3];

		driver_handle.netdev[i] = alloc_etherdev(sizeof(struct igw_priv));
		if (!driver_handle.netdev[i]) {
			pr_err("Could not allocate Network Device%d\n", i);
			err = -ENOMEM;
			return err;
		}

		priv = (struct igw_priv *)netdev_priv(driver_handle.netdev[i]);
		core_idx = 0;
		dev_idx = i + core_idx;
		driver_handle.dev[dev_idx] = &priv[core_idx];

		priv[core_idx].netif_index = i;
		priv[core_idx].netif_priority = i & 0x1;
		priv[core_idx].core_index = (i/2 + core_idx);
		priv[core_idx].__netarch.dev = driver_handle.netdev[i];

		priv[core_idx].__netarch.eth_config.speed_autodetect_target = true;
		priv[core_idx].__netarch.eth_config.speed_target = NET_LINKSPEED_UNKNOWN;
		priv[core_idx].__netarch.eth_config.speed_target_auto = NET_LINKSPEED_UNKNOWN;
		spin_lock_init(&priv->__netarch.lock);

		priv[core_idx].handle = &driver_handle;

		snprintf(drv_name, sizeof(drv_name), "%s%d", IGE_IFNAME, i);
		err = igw_net_ether_setup(driver_handle.netdev[i], drv_name);
		if (err)
			return err;
	}

	interrupt_tasklet_setup(&driver_handle);

	return 0;
}

static int ifrtp_register_device(void)
{
	int i;
	int ret = 0;
	int err = 0;
	int netdev_index;

	for (i = 0; i < driver_handle.netif_index_max; i++) {
		if (!driver_handle.dev[i])
			continue;

		netdev_index = driver_handle.dev[i]->netif_index;
		if (driver_handle.netdev[netdev_index] == NULL)
			continue;

		err = register_netdev(driver_handle.netdev[netdev_index]);

		if (err) {
			pr_err("Failed to register Network Device '%s'\n", driver_handle.netdev[netdev_index]->name);
			ret = err;
		} else {
			pr_info("Network Device '%s' is successfully registerd\n", driver_handle.netdev[netdev_index]->name);
		}
	}
	return ret;
}

static int gmac_sfp_init(void)
{
	ifrtp_mdio_link_reset(0, 0, 0);
	ifrtp_mdio_link_reset(1, 0, 0);
	return 0;
}

static int driver_io_ready(struct platform_device *pdev)
{
	struct IocDiscardInitConfig discard_config;
	int i, ret, page;
	struct igw_priv *dev;

	ret = igw_init_driver();
	if (ret) {
		dev_err(&pdev->dev, "Failed to initialize igw\n");
		return ret;
	}

	ret = ifrtp_i2c_init_pca9555();
	if (ret) {
		dev_err(&pdev->dev, "Failed to initialize machine dependent devices\n");
		return ret;
	}

	ifrtp_irq_disable(&driver_handle);
	ifrtp_irq_clear(&driver_handle);

	discard_config.discard_en = 0xFFFFFFFF;
	dev = driver_handle.dev[0];
	if (!dev) {
		dev_err(&pdev->dev, "no dev %s:300\n", __func__);
		return -EIO;
	}

	ifrtp_set_tx_discard_all(dev,    true);
	ifrtp_set_rx_discard_all(dev,    true);

	page = 0;
	discard_config.page = page;
	discard_config.area_no = 0;
	ifrtp_set_tx_discard(dev, &discard_config);

	discard_config.area_no = 1;
	ifrtp_set_tx_discard(dev, &discard_config);

	discard_config.area_no = 0;
	ifrtp_set_rx_discard(dev, &discard_config);

	discard_config.area_no = 1;
	ifrtp_set_rx_discard(dev, &discard_config);

	ifrtp_enable_stream_filter(dev);

	if (gmac_sfp_init()) {
		dev_err(&pdev->dev, "Failed to start ethernet %d\n", dev->core_index);
		return -EIO;
	}

	ifrtp_transmitter_system_enable(dev);
	ifrtp_receiver_system_enable(dev);

    // TODO may be better to init in other places?
	ret = ifrtp_driver_start();
	if (ret)
		return ret;
	ifrtp_set_rx_filter_enable(dev);

	ret = ifrtp_register_device();
	if (ret)
		return ret;

	ifrtp_set_tx_discard_all(dev, false);
	ifrtp_set_rx_discard_all(dev, false);

	discard_config.discard_en = 0x3F000000;
	for (i = 0; i < 2; ++i) {
		discard_config.area_no = i;
		ifrtp_set_tx_discard(dev, &discard_config);
		ifrtp_set_rx_discard(dev, &discard_config);
	}

	dev_info(&pdev->dev, "IO is successfully initialized\n");

	return 0;

}

void init_fpga(struct igw_driver_handle *handle)
{
	ifrtp_init_network_device_tx(handle->dev[0]);
	ifrtp_init_network_device_rx(handle->dev[0]);
}

int ifrtp_driver_start(void)
{
	init_fpga(&driver_handle);

	ifrtp_irq_enable_raw_packet(driver_handle.dev[0]);

	pr_info("ifrtp_drv is successfully running\n");
	return 0;
}

irqreturn_t igw_tasklet_schedule(void)
{
	irqreturn_t ret = IRQ_NONE;
	bool is_raw_packet_irq_set = false;

	if (ifrtp_irq_get_raw_packet(driver_handle.dev[0])) {
		ifrtp_irq_disable_raw_packet(driver_handle.dev[0]);
		driver_handle.raw_packet_received = true;
		is_raw_packet_irq_set = true;
	}
	if (is_raw_packet_irq_set) {
		tasklet_schedule(&driver_handle.tasklet_raw_packet);
		ret = IRQ_HANDLED;
	}

	return ret;
}

static void ifrtp_release_driver(void)
{
	int i;

	ifrtp_irq_disable(&driver_handle);

	for (i = 0; i < driver_handle.netif_index_max; i++) {
		if (driver_handle.netdev[i]) {
			igw_net_ether_release(driver_handle.netdev[i]);
			free_netdev(driver_handle.netdev[i]);
			driver_handle.netdev[i] = NULL;
		}
	}
}


#define IFRTP_DEVICE_NAME "ifrtp"

static irqreturn_t fpga_interrupt_handler(int irq, void *dev)
{
	irqreturn_t ret = IRQ_NONE;

	if (igw_tasklet_schedule() == IRQ_HANDLED)
		ret = IRQ_HANDLED;

	if (ifrtp_napi_schedule() == IRQ_HANDLED)
		ret = IRQ_HANDLED;

	return ret;
}


int ifrtp_io_init(struct platform_device *pdev)
{
	int ret;
	struct resource *res;

    // init io
	io_remapped = devm_platform_get_and_ioremap_resource(pdev, IFRTP_RES_INDEX_FPGA, &res);
	if (IS_ERR(io_remapped)) {
		dev_err(&pdev->dev, "failed to get and ioremap resource for ifrtp fpga\n");
		return PTR_ERR(io_remapped);
	}
	dev_dbg(&pdev->dev, "0x%llx-0x%llx ioremaped to %px\n", res->start, res->end, io_remapped);

	led_io = devm_platform_get_and_ioremap_resource(pdev, IFRTP_RES_INDEX_LED, &res);
	if (IS_ERR(led_io)) {
		dev_warn(&pdev->dev, "failed to get and ioremap resource for led control: %ld\n",
				PTR_ERR(led_io));
	}
	dev_dbg(&pdev->dev, "0x%llx-0x%llx ioremaped to %px\n", res->start, res->end, led_io);

	ifrtp_shared.irq = platform_get_irq(pdev, 0);
	if (ifrtp_shared.irq > 0) {
		ret = request_irq(ifrtp_shared.irq, fpga_interrupt_handler, 0, IFRTP_DEVICE_NAME, NULL);
		if (ret) {
			dev_err(&pdev->dev, "request_irq(irq: %u) failed.\n", ifrtp_shared.irq);
			return ret;
		}
	} else {
		dev_err(&pdev->dev, "could not get irq\n");
		return ifrtp_shared.irq;
	}

	return driver_io_ready(pdev);
}

/**
 * ifrtp_of_probe(): ifrtp probe function
 * @pdev: Pointer to platform device structure
 *
 * initialization sequence
 *     1. io
 *     2. register value
 *     3. linux network device
 *     4. irq
 *
 * Return: 0,      on success
 *         -ERRNO, on failure
 *
 */
static int ifrtp_probe(struct platform_device *pdev)
{
	int err;

	ifrtp_cfg.driver_handle = &driver_handle;

    // allocate memory and call driver_io_ready()
	err = ifrtp_io_init(pdev);
	if (err) {
		dev_err(&pdev->dev, "failed to init IO. errno : %d\n", err);
		goto igw_init_error;
	}

	err = ifrtp_init_driver(ifrtp_cfg.numvfs);
	if (err) {
		dev_err(&pdev->dev, "ifrtp_init_driver failed: %d", err);
		goto ifrtp_init_error;
	}

	return 0;

ifrtp_init_error:
	ifrtp_cleanup_driver();

igw_init_error:
	ifrtp_release_driver();
	return err;
}

static int ifrtp_remove(struct platform_device *pdev)
{
	ifrtp_cleanup_driver();
	ifrtp_i2c_release_pca9555();
	ifrtp_release_driver();

	if (ifrtp_shared.irq) {
		dev_info(&pdev->dev, "free irq %d\n", ifrtp_shared.irq);
		free_irq(ifrtp_shared.irq, NULL);
	}
	return 0;
}

static const struct of_device_id ifrtp_of_match[] = {
	{ .compatible = "sony,ifrtp", },
	{},
};

MODULE_DEVICE_TABLE(of, ifrtp_of_match);

static struct platform_driver ifrtp_platform_driver = {
	.probe  = ifrtp_probe,
	.remove = ifrtp_remove,
	.driver = {
		.name = KBUILD_MODNAME,
		.of_match_table = ifrtp_of_match,
	}
};

module_platform_driver(ifrtp_platform_driver);

MODULE_DESCRIPTION("Sony ifrtp SFP Ethernet device driver");
MODULE_AUTHOR("Ziyuan Guo <ziyuan.guo@sony.com>");
MODULE_LICENSE("GPL");
MODULE_VERSION("0.1");
