// SPDX-License-Identifier: GPL-2.0
/*
 * Sony ifrtp SFP Ethernet device driver
 *
 * Copyright (c) 2023 Sony Corporation. All rights reserved.
 */
#include <linux/kern_levels.h>
#include <linux/delay.h>
#include "ifrtp.h"
#include "ifrtp_mdio.h"
#include "io.h"
#include "igw_net_address.h"

extern struct ifrtp_config ifrtp_cfg;

int __mdio_write16(uint8_t addr, uint16_t data, int area_no)
{
	// ifrtp_pr_debug("%s(addr=0x%02x, data=0x%04x, area_no=%d)", __func__, addr, data, area_no);
	if (addr > 0x1F) {
		pr_err("invalid mdio address 0x%02x", addr);
		return -1;
	}
	if (area_no > 1) {
		pr_err("invalid mdio area_no %d", area_no);
		return -1;
	}

	unsigned int access_data = MDIO_OP_CL22_WRITE |
							   ((uint32_t)data << MDIO_CL22_DATA_SB) |
							   // ((MDIO_PHY_ADDR_BASE + area_no) << MDIO_PHY_ADDR_SB) |
							   ((MDIO_PHY_ADDR_BASE) << MDIO_PHY_ADDR_SB) |
							   ((uint32_t)addr << MDIO_REG_ADDR_SB) |
							   MDIO_OP_RDY | MDIO_OP_GO;
	unsigned int offset_area_top = OFFSET_NET_GMAC_TOP + OFFSET_NET_GMAC_CTL_TOP +
								   (area_no * OFFSET_NET_GMAC_CTL_SIZE);
	unsigned int address = offset_area_top + OFFSET_NET_GMAC_CTL_MDIO_ACCESS;
    // TODO (guo) check if it's necessery to add a lock
	ifrtp_iow32(NULL, access_data, address);
    // TODO (guo) check if it's necessery to wait for mdio ready
	return 0;
}

int __mdio_read16(uint8_t addr, uint16_t *data, int area_no)
{
	// ifrtp_pr_debug("%s(addr=0x%02x, area_no=%d)", __func__, addr, area_no);
	if (addr > 0x1F) {
		pr_err("invalid mdio address 0x%02x", addr);
		return -1;
	}
	if (area_no > 1) {
		pr_err("invalid mdio area_no %d", area_no);
		return -1;
	}

	unsigned int access_data = MDIO_OP_CL22_READ |
							   // ((MDIO_PHY_ADDR_BASE + area_no) << MDIO_PHY_ADDR_SB) |
							   ((MDIO_PHY_ADDR_BASE) << MDIO_PHY_ADDR_SB) |
							   ((uint32_t)addr << MDIO_REG_ADDR_SB) |
							   MDIO_OP_RDY | MDIO_OP_GO;
	unsigned int offset_area_top = OFFSET_NET_GMAC_TOP + OFFSET_NET_GMAC_CTL_TOP +
								   (area_no * OFFSET_NET_GMAC_CTL_SIZE);
	unsigned int address = offset_area_top + OFFSET_NET_GMAC_CTL_MDIO_ACCESS;

    // TODO (guo) check if it's necessery to add a lock
	ifrtp_iow32(NULL, access_data, address);
	uint32_t res;

	while (!((res = ifrtp_ior32(NULL, address)) & MDIO_OP_RDY)) {
		// usleep(1);
	}
	*data = (res & MDIO_CL22_DATA_MASK) >> MDIO_CL22_DATA_SB;

	ifrtp_pr_debug("%s(addr=0x%02x, data=0x%04x, area_no=%d)", __func__, addr, *data, area_no);
	return 0;
}


// refer to https://docs.xilinx.com/v/u/en-US/gig_eth_pcs_pma_ug155 12.2 Management Registers

bool ifrtp_mdio_sfp_is_linked_up(int interface)
{
	uint16_t data;

	// ifrtp_pr_debug("mdio_sfp_is_linked_up(interface=%d)", interface);
    // LinkStatus need to be read twice
	__mdio_read16(1, &data, interface);

	if (__mdio_read16(0, &data, interface))
		return false;

	if ((data & 0x0140) != 0x0140)
		return false;

	if (__mdio_read16(1, &data, interface))
		return false;

	if (data & 4)
		return true;
	else
		return false;
}

int ifrtp_mdio_link_reset(int interface, int timeout_ms, int retry)
{
	int i;

	bool autoneg = ifrtp_cfg.driver_handle->dev[interface]->__netarch.eth_config.speed_autodetect_target;

	pr_info("mdio reset link %d\n", interface);

    // Core reset
	__mdio_write16(0, 0x8000, interface);

	for (i = 0; i <= retry; ++i) {
		// Auto-Negotiation Enable
		// Auto-Negotiation Restart
		// Full-Duplex Mode
		// Speed Selection 10 (1000Mbps)
		__mdio_write16(0, autoneg ? 0x1340 : 0x0140, interface);
		if (timeout_ms) {
			msleep_interruptible(timeout_ms);
			if (ifrtp_mdio_sfp_is_linked_up(interface))
				return 0;
		}
	}
	return -1;
}

void ifrtp_mdio_link_power_down(int interface)
{
	__mdio_write16(0, 0x0800, interface);
}
