// SPDX-License-Identifier: GPL-2.0
/*
 * Sony ifrtp SFP Ethernet device driver
 *
 * Copyright (c) 2023 Sony Corporation. All rights reserved.
 */
#include <linux/version.h>
#include <linux/netdevice.h>
#include <linux/if_ether.h>
#include <linux/etherdevice.h>
#include <linux/ethtool.h>
#include <linux/byteorder/generic.h>

#include "ifrtp_mdio.h"
#include "igw_driver.h"
#include "igw_netreg.h"
#include "igw_net_address.h"

#include "ifrtp.h"
#include "ifrtp_filter.h"
#include "io.h"


#define IGW_NET_LINKDETECTION_TIMER_TIMEOUT    (HZ >> 1) //500msec
#define IGW_NET_MAX_PACKET_SIZE (1024*64)

static void igw_timer_handler(struct timer_list *t);

static bool check_packet_receive(struct igw_priv *dev, bool is_init);

#define LINK_STATE_DOWN                 0x00
#define LINK_STATE_UNKNOWN              0x01
#define LINK_STATE_UP                   0x02
#define LINK_STATE_UP_PACKET_RECEIVED   0x04

#define IGW_HEADER_WORDS     4
#define IGW_HEADER_BYTES     (IGW_HEADER_WORDS << 2)
#define IGW_HEADER_TYPE_MASK 0x00FF0000
#define IGW_HEADER_SIZE_MASK 0x0000FFFF
#define IGW_HEADER_TYPE_SB   16
#define IGW_HEADER_MAC_MASK  0xFF000000
#define IGW_HEADER_MAC_SB    24


static bool is_open_tx_discard[IFRTP_PHY_COUNT];
static bool is_open_rx_discard[IFRTP_PHY_COUNT];
static bool is_open_mii_filter[IFRTP_PHY_COUNT];

static void open_tx_discard(struct net_device *dev)
{
	struct igw_priv *priv = (struct igw_priv *) netdev_priv(dev);

	if (is_open_tx_discard[priv->netif_index]) {
		return;
	}

	spin_lock(&priv->__netarch.lock);

	ifrtp_set_tx_discard_all(priv, false);
	ifrtp_set_tx_discard_by_devindex(priv, false);

	is_open_tx_discard[priv->netif_index] = true;

	spin_unlock(&priv->__netarch.lock);
}

static void open_rx_discard(struct net_device *dev)
{
	struct igw_priv *priv = (struct igw_priv *) netdev_priv(dev);

	if (is_open_rx_discard[priv->netif_index]) {
		return;
	}

	spin_lock(&priv->__netarch.lock);

	ifrtp_set_rx_discard_all(priv, false);
	ifrtp_set_rx_discard_by_devindex(priv, false);

	is_open_rx_discard[priv->netif_index] = true;

	spin_unlock(&priv->__netarch.lock);
}

static void open_mii_filter(struct net_device *dev)
{
	struct igw_priv *priv = (struct igw_priv *) netdev_priv(dev);

	if (is_open_mii_filter[priv->netif_index]) {
		return;
	}

	spin_lock(&priv->__netarch.lock);

    // packet受信検知できるように初期化(現状値を取得)しておく.
	check_packet_receive(priv, true);

    // TODO ? enable ethernet mii path

	is_open_mii_filter[priv->netif_index] = true;

	spin_unlock(&priv->__netarch.lock);
}

static void close_mii_filter(struct net_device *dev)
{
	struct igw_priv *priv = (struct igw_priv *) netdev_priv(dev);

	if (!is_open_mii_filter[priv->netif_index]) {
		return;
	}

    // TODO ? disable ethernet mii path

	is_open_mii_filter[priv->netif_index] = false;
}

static int igw_open(struct net_device *dev)
{
	struct igw_priv *priv = (struct igw_priv *) netdev_priv(dev);
	static bool initilize[IFRTP_PHY_COUNT];

    // 最初の openの場合のみadd_timerして、ポーリング関数を呼び出すようにする.
	if (!initilize[priv->netif_index]) {
		add_timer(&priv->__netarch.link_check_timer);
		initilize[priv->netif_index] = true;
	}

	return 0;
}

static int igw_close(struct net_device *dev)
{
	netif_stop_queue(dev);
	pr_debug("%s netif stop queuing\n", dev->name);

    // mdio_link_power_down?
    // close_discard?
    // cannot bring up the interface again if close_discard
    // TODO (guo): turn off phy properly and fix ifrtp_open

	return 0;
}

static struct net_device_stats *igw_get_stats(struct net_device *dev)
{
	struct igw_priv *priv;

	priv = (struct igw_priv *) netdev_priv(dev);

	return &priv->__netarch.stats;
}

static int igw_mactype_to_devindex(unsigned int src_mactype, int *dest_devindex)
{
	int index = 0;

	switch (src_mactype) {
	case 4:
		index = 0;
		break;
	case 5:
		index = 1;
		break;
	default:
		return -EINVAL;
	}
	(*dest_devindex) = index;
	return 0;
}

static int igw_devindex_to_mactype(int src_index, unsigned int *dest_mactype)
{
	unsigned int mactype = 0;

	switch (src_index) {
	case 0:
		mactype = 4;
		break;
	case 1:
		mactype = 5;
		break;
	default:
		return -EINVAL;
	}
	(*dest_mactype) = mactype;
	return 0;
}

static DEFINE_SPINLOCK(transmitter_lock);
#define OUTPUT_TIMEOUT  1000
#define OUTPUT_SIZE_MAX 0x000007F0

static int transmitter_system_send(struct igw_priv *dev, unsigned int *header, struct sk_buff *data)
{
	int i = 0;
	int offset = 0;

	int payload_real_size    = header[0] & IGW_HEADER_SIZE_MASK;
	int payload_aligned_size = (payload_real_size + 15) & ~15;  /* Internal Outputへの書き込みは16byte align */

	unsigned int *source;
	unsigned int timeout = 0;
	unsigned long flags;

	unsigned char data_with_padding[60];

	if (payload_aligned_size > OUTPUT_SIZE_MAX)
		return -EMSGSIZE;

	while (ifrtp_ior32(dev, OFFSET_NETTX_TRANS_SYS_TOP + OFFSET_NETTX_TRANS_SYS_QUEUE_SPACE) == 0 ||
			payload_aligned_size + IGW_HEADER_BYTES  > ifrtp_ior32(dev, OFFSET_NETTX_TRANS_SYS_TOP + OFFSET_NETTX_TRANS_SYS_BUFFER_SPACE)) {
		timeout++;
		if (timeout > OUTPUT_TIMEOUT)
			return -EBUSY;
	}

	spin_lock_irqsave(&transmitter_lock, flags);

	if (data->len < 60) {
		memset(data_with_padding, 0x0, sizeof(data_with_padding));
		memcpy(data_with_padding, data->data, data->len);
		source = (unsigned int *)data_with_padding;
	} else {
		source = (unsigned int *)data->data;
	}

	offset = ifrtp_ior32(dev, OFFSET_NETTX_TRANS_SYS_TOP + OFFSET_NETTX_TRANS_SYS_BUFFER_OFFSET);

	ifrtp_memcpy_toio(dev, header, offset + OFFSET_NETTX_TRANS_SYS_TOP + OFFSET_NETTX_TRANS_SYS_BUFFER, IGW_HEADER_WORDS);
	offset += IGW_HEADER_WORDS * sizeof(unsigned int);


    /* Payloadの書き込み */
	for (i = 0; i < payload_aligned_size / sizeof(unsigned int); i++) {
		ifrtp_iow32(dev, htonl(source[i]), offset + OFFSET_NETTX_TRANS_SYS_TOP + OFFSET_NETTX_TRANS_SYS_BUFFER);
		offset += sizeof(unsigned int);
	}
    // TODO let FPGA do htonl for us

    /* Check-In */
	ifrtp_iow32(dev, payload_aligned_size + IGW_HEADER_BYTES, OFFSET_NETTX_TRANS_SYS_TOP + OFFSET_NETTX_TRANS_SYS_BUFFER_UPDATE);

	spin_unlock_irqrestore(&transmitter_lock, flags);


	return 0;
}

static int igw_send(struct sk_buff *skb, struct net_device *dev)
{
	int result = 0;
	unsigned int len = 0;

	struct igw_priv *priv;
	unsigned int header[IGW_HEADER_WORDS];
	unsigned int type = 0;
	unsigned int mactype = 0;

	priv = (struct igw_priv *) netdev_priv(dev);

	type = TX_TYPE_MAIN + (priv->netif_priority);

	len = skb->len;
	if (len <= 0 || skb->data == NULL) {
		dev_kfree_skb_any(skb);
		return NETDEV_TX_OK;
	}

    // padding will be performed in transmitter_system_send()
	if (len < 60) {
		len = 60;
	}

	if (ntohs(skb->protocol) == ETH_P_IP && ip_hdr(skb)->protocol == SOL_UDP) {
		ifrtp_filter_accept_ephemeral_port(priv, ntohs(udp_hdr(skb)->source));
	}

	if (igw_devindex_to_mactype(priv->netif_priority, &mactype)) {
		result = -EINVAL;
		goto IGW_SEND_FIN;
	}
	header[0] = ((type << IGW_HEADER_TYPE_SB) & IGW_HEADER_TYPE_MASK) | (len & IGW_HEADER_SIZE_MASK);
	header[1] = (mactype << IGW_HEADER_MAC_SB) & IGW_HEADER_MAC_MASK;
	header[2] = 0;
	header[3] = 0;

	netif_trans_update(dev);
	result = transmitter_system_send(priv, header, skb);


IGW_SEND_FIN:

    // 統計情報の更新
	if (result == 0) {
		priv->__netarch.stats.tx_packets++;
		priv->__netarch.stats.tx_bytes += skb->len;
	} else {
		pr_warn("%s failed to send packet(code:%d)\n", dev->name, result);
		priv->__netarch.stats.tx_errors++;
		priv->__netarch.stats.tx_dropped++;
	}

	dev_kfree_skb_any(skb);
	return NETDEV_TX_OK;
}

static int igw_set_mac_address(struct net_device *dev, void *addr)
{
	struct sockaddr *mac = (struct sockaddr *) addr;

	memcpy(dev->dev_addr, mac->sa_data, IFHWADDRLEN);
	pr_debug("%s set mac %02x:%02x:%02x:%02x:%02x:%02x\n",
			dev->name,
			dev->dev_addr[0],
			dev->dev_addr[1],
			dev->dev_addr[2],
			dev->dev_addr[3],
			dev->dev_addr[4],
			dev->dev_addr[5]
	);

	struct IocNetTxSourceMacAddressConfig mac_config;

	memcpy(mac_config.eth_src, mac->sa_data, sizeof(mac_config.eth_src));

	if (memcmp(dev->name, IGE_IFNAME"0", sizeof(IGE_IFNAME)) == 0) {
		ifrtp_set_tx_source_primary_mac_address(netdev_priv(dev), &mac_config);
	} else {
		ifrtp_set_tx_source_secondary_mac_address(netdev_priv(dev), &mac_config);
	}

	return 0;
}

static inline int ring_buff_offset(int offset)
{
	return offset & OFFSET_RECV_SYSTEM_BUFFER_ADDR_MASK;
}

static int igw_receiver_system_get_level(struct igw_priv *dev)
{
	return ifrtp_ior32(dev, OFFSET_RECV_SYSTEM_TOP + OFFSET_RECV_SYSTEM_QUEUE_LEVEL);
}


static int igw_receiver_system_receive_header(struct igw_priv *dev, unsigned int *header)
{
	int i = 0;
	int offset = 0;

    /* Check packet length */
	if (ifrtp_ior32(dev, OFFSET_RECV_SYSTEM_TOP + OFFSET_RECV_SYSTEM_QUEUE_LEVEL) == 0) {
		return 0;
	}

	offset = ifrtp_ior32(dev, OFFSET_RECV_SYSTEM_TOP + OFFSET_RECV_SYSTEM_BUFFER_OFFSET);

    /* Headerの取得 */
	for (i = 0; i < IGW_HEADER_WORDS; i++) {
		offset = ring_buff_offset(offset);
		header[i] = ifrtp_ior32(dev, offset + OFFSET_RECV_SYSTEM_TOP + OFFSET_RECV_SYSTEM_BUFFER);
		offset += sizeof(unsigned int);
//        pr_info("[ifrtp_reciever_system_receive_header]header[%0d]=%08x",i,header[i]);
	}

	return IGW_HEADER_BYTES;
}

static int igw_receiver_system_receive_payload(struct igw_priv *dev, void *data, unsigned int size_on_header)
{
	int i = 0;
	unsigned int  payload_size   = 0;
	int offset = 0;

	unsigned int *destination = (unsigned int *)data;

    /* Check packet length */
	if (ifrtp_ior32(dev, OFFSET_RECV_SYSTEM_TOP + OFFSET_RECV_SYSTEM_QUEUE_LEVEL) == 0) {
		return 0;
	}

	payload_size   = ifrtp_ior32(dev, OFFSET_RECV_SYSTEM_TOP + OFFSET_RECV_SYSTEM_BUFFER_LEVEL)  - IGW_HEADER_BYTES;
	offset = ifrtp_ior32(dev, OFFSET_RECV_SYSTEM_TOP + OFFSET_RECV_SYSTEM_BUFFER_OFFSET) + IGW_HEADER_BYTES;

	if (payload_size - size_on_header >= 16) {
		pr_err("[%s] Payload size %d is too larger than size:%d on header.\n", __func__, payload_size, size_on_header);
		//headerに明記されているpaketサイズがData部に詰められるので問題ない
		//よって、returnせず以降の処理を実施する
	}

	if (size_on_header > payload_size) {
		//Payloadにゴミが詰め込まれる可能性があるためreturnする
		pr_err("[%s] size:%d on header is larger than Payload size:%d.\n", __func__, size_on_header, payload_size);
		return 0;
	}

    /* Payloadの取得 */
	for (i = 0; i <= size_on_header / sizeof(unsigned int); i++) {
		offset = ring_buff_offset(offset);
		destination[i] = ntohl(ifrtp_ior32(dev, offset + OFFSET_RECV_SYSTEM_TOP + OFFSET_RECV_SYSTEM_BUFFER));
		offset += sizeof(unsigned int);
	}

    /* Check-Out */
	ifrtp_iow32(dev, 0, OFFSET_RECV_SYSTEM_TOP + OFFSET_RECV_SYSTEM_BUFFER_UPDATE);

	return size_on_header;
}


static void igw_receiver_system_force_checkout(struct igw_priv *dev)
{
    /* Check-Out */
	ifrtp_iow32(dev, 0, OFFSET_RECV_SYSTEM_TOP + OFFSET_RECV_SYSTEM_BUFFER_UPDATE);
}

static void igw_receive(struct igw_priv *priv)
{
	struct net_device *netdev;
	struct igw_priv *dev[2];
	unsigned int packet_size = 0;

	struct sk_buff *skb;
	unsigned int header[IGW_HEADER_WORDS];
	unsigned int type = 0;
	unsigned int size_on_header = 0;
	unsigned int mactype = 0;
	int devindex = 0;
	int packet_date_size;

	if (priv->netif_priority > 0) {
		//SecondaryのNetIFで割込を受けることはない
		pr_debug("%s(): netif_prioirty %d\n", __func__, priv->netif_priority);
		return;
	}

	dev[0] = priv;
	dev[1] = priv->handle->dev[priv->netif_index + 1]; //SecondaryInterfaceの参照
    // InternalInputが空になるまでPacket受信処理
	while ((igw_receiver_system_get_level(priv) > 0) && (packet_size < IGW_NET_MAX_PACKET_SIZE)) {
		// Headerの取得
		igw_receiver_system_receive_header(priv, header);

		// HeaderからType/Sirrzeの取得
		type = (header[0] & IGW_HEADER_TYPE_MASK) >> IGW_HEADER_TYPE_SB;
		size_on_header = (header[0] & IGW_HEADER_SIZE_MASK);
		mactype = (header[1] & IGW_HEADER_MAC_MASK) >> IGW_HEADER_MAC_SB;

		if (igw_mactype_to_devindex(mactype, &devindex)) {
			pr_warn_ratelimited("Packet is received from unknown MAC index:%d\n", type);

			// 強制Checkout
			igw_receiver_system_force_checkout(priv);
		} else {
			// Primary/Seconderyどちらのインターフェース受信したかで参照するデバイス構造体を切り替える
			netdev = dev[devindex]->__netarch.dev;
			if (netdev == NULL) {
				// 強制Checkout
				igw_receiver_system_force_checkout(priv);
			} else {
				if (!(netif_carrier_ok(netdev) && netif_running(netdev))) {
					// ifconfig upしていない場合 or LinkDownしている場合は, Packetを破棄(強制Checkout)
					igw_receiver_system_force_checkout(priv);
				} else {
					// Socket BufferのAllocate
					skb = dev_alloc_skb(size_on_header + 4);
					if (!skb) {
						// packetを破棄
						igw_receiver_system_force_checkout(priv);
						dev[devindex]->__netarch.stats.rx_dropped++;
						pr_err_ratelimited("%s Could not allocate receive buffer\n", netdev->name);
					} else {
						// Internal Inputからパケットを取得
						packet_date_size = igw_receiver_system_receive_payload(priv, skb_put(skb, size_on_header), size_on_header);

						if (packet_date_size == 0) {
							igw_receiver_system_force_checkout(priv);
							dev[devindex]->__netarch.stats.rx_dropped++;
							dev_kfree_skb_any(skb);
							continue;
						}

						// TODO: Checksum演算結果の判定処理

						// Socket Bufferパラメータをセット
						skb->dev = netdev;
						skb->protocol = eth_type_trans(skb, netdev);
						skb->ip_summed = CHECKSUM_NONE;

						// Kernelにパケットを送信
						if (netif_rx(skb) == NET_RX_SUCCESS) {
							// 統計情報の更新
							dev[devindex]->__netarch.stats.rx_packets++;
							dev[devindex]->__netarch.stats.rx_bytes += size_on_header;
						} else {
							dev[devindex]->__netarch.stats.rx_dropped++;
						}
						packet_size += size_on_header;
					}
				}
			}
		}
	}
	return;
}

static int igw_get_link_ksettings(struct net_device *dev, struct ethtool_link_ksettings *elk)
{
	struct ethtool_link_settings *const ecmd = &elk->base;
	int result = 0;
	struct igw_priv *priv = (struct igw_priv *) netdev_priv(dev);
	bool is_link_up = netif_carrier_ok(dev);

	spin_lock(&priv->__netarch.lock);

	ecmd->transceiver = XCVR_EXTERNAL;
	ecmd->port = PORT_FIBRE;
	ecmd->autoneg = (__u8)(priv->__netarch.eth_config.speed_autodetect_target ? AUTONEG_ENABLE : AUTONEG_DISABLE);
	ecmd->duplex = (is_link_up ? DUPLEX_FULL : DUPLEX_UNKNOWN);
    // upしていれば、serdes rate, Downなら unknown.

	ethtool_link_ksettings_add_link_mode(elk, supported, FIBRE);
    // TODO add speed detect
	ecmd->speed = (__u32)(is_link_up ? NET_LINKSPEED_1G : SPEED_UNKNOWN);

	spin_unlock(&priv->__netarch.lock);

	return result;
}

static int igw_set_link_ksettings(struct net_device *dev, const struct ethtool_link_ksettings *elk)
{
	const struct ethtool_link_settings *const ecmd = &elk->base;
	int result = 0;
	struct igw_priv *priv = (struct igw_priv *) netdev_priv(dev);

	spin_lock(&priv->__netarch.lock);

	if (ecmd->autoneg != priv->__netarch.eth_config.speed_autodetect_target) {
		switch (ecmd->autoneg) {
		case 1:
			priv->__netarch.eth_config.speed_autodetect_target = true;
			break;
		case 0:
			priv->__netarch.eth_config.speed_autodetect_target = false;
			break;
		default:
			netdev_err(dev, "Could not set autodetect to value %d from ethtool\n", ecmd->autoneg);
			result = -EINVAL;
		}
	}

	if (!priv->__netarch.eth_config.speed_autodetect_target) {
		if (ecmd->speed != priv->__netarch.eth_config.speed_target && ecmd->speed != (__u16)SPEED_UNKNOWN) {

			switch (ecmd->speed) {
				case NET_LINKSPEED_1G:
					priv->__netarch.eth_config.speed_target = NET_LINKSPEED_1G;
					break;
				default:
					netdev_err(dev, "Failed to set ethernet speed to %dMbps, %dMbps is not supported\n", (int)ecmd->speed, (int)ecmd->speed);
					result = -EINVAL;
			}
		}
	}

	ifrtp_mdio_link_reset(priv->netif_priority, 0, 0);

	spin_unlock(&priv->__netarch.lock);
	return result;
}


static void igw_timer_handler(struct timer_list *t)
{
	struct igw_priv *priv = from_timer(priv, t, __netarch.link_check_timer);

	schedule_work((struct work_struct *)priv->__netarch.link_check_queue_work);

	mod_timer(&priv->__netarch.link_check_timer, jiffies + IGW_NET_LINKDETECTION_TIMER_TIMEOUT);
}


static bool check_packet_receive(struct igw_priv *dev, bool is_init)
{
	const unsigned int TIME_OUT_NO_PACKET_COUNT = 10;
	static unsigned int prev_recv_good_frame_cnt[IFRTP_PHY_COUNT];
	static unsigned int no_packet_count[IFRTP_PHY_COUNT];
	bool ret = is_init;
	unsigned int recv_good_frame_cnt = ifrtp_get_rx_xgmac_good_frame_lower(dev);

	if (is_init) {
		no_packet_count[dev->netif_index] = 0;
	}
    // 初期化でなくgoodframeが増えたかどうか.
	else if (recv_good_frame_cnt != prev_recv_good_frame_cnt[dev->netif_index]) {
		ret = true;
	}
    // goodframe増えない.
	else {
		no_packet_count[dev->netif_index]++;
		if (no_packet_count[dev->netif_index] == TIME_OUT_NO_PACKET_COUNT) {// timeout.
			pr_err("IF[%d] Timed out with no packets received, interface is going to up.\n", dev->netif_index);
			ret = true;
		}
	}
	prev_recv_good_frame_cnt[dev->netif_index] = recv_good_frame_cnt;
	return ret;
}

static unsigned int get_phy_link_state(struct igw_priv *dev, unsigned int prev_phy_link_state)
{
	bool is_sfp_linked = ifrtp_sfp_is_linked_up(dev);
	bool is_serdes_linked = true;

	if (is_sfp_linked && is_serdes_linked) {
		return LINK_STATE_UP |
			(prev_phy_link_state & LINK_STATE_UP_PACKET_RECEIVED);
	}
	return LINK_STATE_DOWN;
}

static void igw_link_check_handler(struct work_struct *work)
{
	static unsigned int prev_phy_link_state[IFRTP_PHY_COUNT];

	struct ifrtp_link_check_queue_work *priv_work = (struct ifrtp_link_check_queue_work *)work;
	struct igw_priv *priv = priv_work->dev;
	unsigned int phy_link_state = get_phy_link_state(priv, prev_phy_link_state[priv->netif_index]);
	bool is_serdes_change_config = false;

	bool run = netif_running(priv->__netarch.dev);

	if (phy_link_state & LINK_STATE_UP && priv->__netarch.dev->flags & IFF_UP) {
		ifrtp_led_on(priv->netif_priority);
	} else {
		ifrtp_led_off(priv->netif_priority);
	}

    // UNKNOWN or 状態 or 設定 に変化があったどうか.
	if (phy_link_state & LINK_STATE_UNKNOWN ||
		phy_link_state != prev_phy_link_state[priv->netif_index] ||
		is_serdes_change_config
		) {
		// open中.
		if (run) {
			if (!netif_queue_stopped(priv->__netarch.dev)) {
				netif_stop_queue(priv->__netarch.dev);
				pr_info("%s netif stop queuing\n", priv->__netarch.dev->name);
			}
		}

		// 遮断させる.
		close_mii_filter(priv->__netarch.dev);
		// close_discard?

		// Downさせる
		if (netif_carrier_ok(priv->__netarch.dev)) {
			netif_carrier_off(priv->__netarch.dev);
		}

		// serdesの構築(変更がなくてもEthrer Resetは実施させる).
		// link stateの更新.
		phy_link_state = get_phy_link_state(priv, phy_link_state);

		if (phy_link_state & LINK_STATE_UP) {
			// Upさせる.
			netif_carrier_on(priv->__netarch.dev);
			pr_info("IF[%d] Up\n", priv->netif_index);

			// mii filterまではrunning関係なく開けておく.
			open_mii_filter(priv->__netarch.dev);

			if (run) {
				// packetをkernelが送受信できる状態にする(RxはXGMACまで開通).
				open_tx_discard(priv->__netarch.dev);
				netif_wake_queue(priv->__netarch.dev);
				pr_info("%s netif wake queuing\n", priv->__netarch.dev->name);
			}
		} else if (phy_link_state & LINK_STATE_UNKNOWN) {
			if (!(prev_phy_link_state[priv->netif_index] & LINK_STATE_UNKNOWN)) { // 前回がUnKnown以外の場合.
				pr_info("IF[%d] Link Unknown\n", priv->netif_index);
			}
		} else if (phy_link_state == LINK_STATE_DOWN) {
			if (!(prev_phy_link_state[priv->netif_index] & LINK_STATE_DOWN)) { // 前回がDown以外の場合.
				pr_info("IF[%d] Down\n", priv->netif_index);
			}
		}
	}
    // Up.
	else if (phy_link_state & LINK_STATE_UP) {

		if (!(phy_link_state & LINK_STATE_UP_PACKET_RECEIVED)) { //パケット受信まだ.
			// パケットが来たかどうかのチェック.
			if (check_packet_receive(priv, false)) {
				// ハマり対策リセット.
				// TODO? XGMAC Reset.
				// Unpack Reset.
				ifrtp_reset_unpack(priv);

				phy_link_state |= LINK_STATE_UP_PACKET_RECEIVED;// 受信できたフラグを付与.
				pr_info("IF[%d] Up Packet Received\n", priv->netif_index);
			}
		}

		if (run) {
			if (phy_link_state & LINK_STATE_UP_PACKET_RECEIVED) {
				open_rx_discard(priv->__netarch.dev);
			}

			if (netif_queue_stopped(priv->__netarch.dev)) {
				netif_wake_queue(priv->__netarch.dev);
				pr_info("%s netif wake queuing\n", priv->__netarch.dev->name);
			}
		}
	}

    //前回状態の更新.
	prev_phy_link_state[priv->netif_index] = phy_link_state;

}

static const struct net_device_ops igw_dev_ops = {
		.ndo_open = igw_open,
		.ndo_stop = igw_close,
		.ndo_get_stats = igw_get_stats,
		.ndo_start_xmit = igw_send,
		.ndo_set_mac_address = igw_set_mac_address,
};

static const struct ethtool_ops igw_ethtool_ops = {
		.get_link_ksettings = igw_get_link_ksettings,
		.set_link_ksettings = igw_set_link_ksettings,
};

int igw_net_ether_setup(struct net_device *dev, const char *drv_name)
{
	struct igw_priv *priv;

	if (dev == NULL) {
		return -EINVAL;
	}

	ether_setup(dev);
	dev->netdev_ops = &igw_dev_ops;
	dev->watchdog_timeo = IGW_NET_LINKDETECTION_TIMER_TIMEOUT;
	netif_carrier_off(dev);
	snprintf(dev->name, sizeof(dev->name), drv_name);
	dev->ethtool_ops = &igw_ethtool_ops;

	priv = (struct igw_priv *)netdev_priv(dev);
	if (priv == NULL) {
		return -EINVAL;
	}

	int err = ifrtp_i2c_read_mac_addr(priv->netif_priority, dev->dev_addr);

	if (err < 0) {
		pr_err("Error Reading MAC Address: %d\n", err);
		return err;
	}
	if (!is_valid_ether_addr(dev->dev_addr)) {
		pr_err("Invalid MAC Address: %pM\n", dev->dev_addr);
	}

	struct IocNetTxSourceMacAddressConfig mac_config;

	memcpy(mac_config.eth_src, dev->dev_addr, sizeof(mac_config.eth_src));
	if (priv->netif_priority) {
		ifrtp_set_tx_source_secondary_mac_address(NULL, &mac_config);
	} else {
		ifrtp_set_tx_source_primary_mac_address(NULL, &mac_config);
	}

	timer_setup(&priv->__netarch.link_check_timer, igw_timer_handler, 0);
	priv->__netarch.link_check_timer.expires = jiffies + IGW_NET_LINKDETECTION_TIMER_TIMEOUT;

	priv->__netarch.link_check_queue_work = kmalloc(sizeof(struct ifrtp_link_check_queue_work), GFP_KERNEL);
	if (priv->__netarch.link_check_queue_work == NULL) {
		return -ENOMEM;
	}
	priv->__netarch.link_check_queue_work->dev = priv;
	INIT_WORK((struct work_struct *)(priv->__netarch.link_check_queue_work), igw_link_check_handler);
	if (get_phy_link_state(priv, 0) == LINK_STATE_UP)
		netif_carrier_on(priv->__netarch.dev);

	return 0;
}

int igw_net_ether_release(struct net_device *dev)
{
	int ret = 0;
	struct igw_priv *priv;

	ret = igw_close(dev);

	priv = (struct igw_priv *)netdev_priv(dev);
	del_timer(&priv->__netarch.link_check_timer);

	flush_scheduled_work();
	kfree(priv->__netarch.link_check_queue_work);

	unregister_netdev(dev);
	return ret;
}

int igw_net_raw_packet_isr(struct igw_priv *dev)
{
	igw_receive(dev);
	return 0;
}

