// SPDX-License-Identifier: GPL-2.0
/*
 * Sony ifrtp SFP Ethernet device driver
 *
 * Copyright (c) 2023 Sony Corporation. All rights reserved.
 */
#include "ifrtp.h"
#include "io.h"
#include "igw_net_address.h"
#include "igw_netreg.h"

//--------------------------------------------------
// Unpack
//--------------------------------------------------
int ifrtp_set_tx_pkt_unpack(struct igw_priv *dev, struct IocPktUnpackTxDirInitConfig *config)
{
	unsigned int offset  = 0x0;
	unsigned int address = 0x0;

	offset = OFFSET_NETTX_PKT_UNPACK_TOP + OFFSET_NETTX_PKT_UNPACK_EN;

	address = offset;
	ifrtp_iow32(dev, config->unpack_enable, address);

	return 0;
}

//----------
int ifrtp_set_rx_pkt_unpack(struct igw_priv *dev, struct IocPktUnpackRxInitConfig *config)
{
	int result  = 0;
	int area_no = 0;
	unsigned int offset_area_top = 0x0;
	unsigned int address         = 0x0;

	area_no = config->area_no;

	if (area_no >= 0 && area_no < OFFSET_NETRX_UNPACK_CTL_MAX) {
		offset_area_top = OFFSET_NETRX_UNPACK_TOP + OFFSET_NETRX_UNPACK_CTL_TOP + (area_no * OFFSET_NETRX_UNPACK_CTL_SIZE);

		address = offset_area_top + OFFSET_NETRX_UNPACK_CTL_UNPACKED_TYPE_CNT;
		ifrtp_iow32(dev, config->unpack_every_type, address);

	} else {
		result = -EFAULT;
	}

	return result;
}

// TODO (guo) check if this is still needed
int ifrtp_reset_unpack(struct igw_priv *dev)
{
	const unsigned int addr =
		OFFSET_NETRX_UNPACK_TOP +
		OFFSET_NETRX_UNPACK_CTL_TOP + (OFFSET_NETRX_UNPACK_CTL_SIZE * dev->netif_priority) +
		OFFSET_NETRX_UNPACK_CTL_RESET;
	ifrtp_iow32(dev, 0x1, addr);
	usleep_range(200, 300);
	ifrtp_iow32(dev, 0x0, addr);

	return 0;
}

//--------------------------------------------------
// Pack
//--------------------------------------------------
int ifrtp_set_tx_pkt_pack(struct igw_priv *dev, struct IocPktPackTxDirInitConfig *config)
{
	int result = 0;
	int packet_pack_no = 0;
	unsigned int offset_channel_top  = 0x0;
	unsigned int address             = 0x0;

	packet_pack_no = config->packet_pack_no;

	if (packet_pack_no >= 0 && packet_pack_no < OFFSET_NETTX_PKT_PACK_MAX) {
		offset_channel_top = OFFSET_NETTX_PKT_PACK_TOP + OFFSET_NETTX_PKT_PACK_EN_TOP + (packet_pack_no * OFFSET_NETTX_PKT_PACK_SIZE);

		address = offset_channel_top + OFFSET_NETTX_PKT_PACK_EN_ENABLE;
		ifrtp_iow32(dev, config->pack_enable, address);
	} else {
		result = -EFAULT;
	}

	return result;
}

int ifrtp_set_rx_pkt_pack(struct igw_priv *dev, struct IocPktPackRxInitConfig *config)
{
	int result = 0;
	int area_no = 0;
	unsigned int offset_area_top = 0x0;
	unsigned int address         = 0x0;

	area_no = config->area_no;

	if (area_no >= 0 && area_no < OFFSET_NETRX_PKT_PACK_CTL_MAX) {
		offset_area_top = OFFSET_NETRX_PKT_PACK_TOP + OFFSET_NETRX_PKT_PACK_CTL_TOP + (area_no * OFFSET_NETRX_PKT_PACK_CTL_SISE);

		address = offset_area_top + OFFSET_NETRX_PKT_PACK_CTL_PACK_EN;
		ifrtp_iow32(dev, config->no_packing, address);
	} else {
		result = -EFAULT;
	}

	return result;
}

//--------------------------------------------------
// Discard
//--------------------------------------------------
static DEFINE_SPINLOCK(discard_tx_lock);
static DEFINE_SPINLOCK(discard_rx_lock);

static inline bool get_discard_tx_base_addr(u32 *base_addr, int area_no)
{
	if ((area_no < 0) || (area_no >= OFFSET_NETTX_DISCARD_CTL_MAX)) {
		return false;
	}

	*base_addr = OFFSET_NETTX_DISCARD_TOP + OFFSET_NETTX_DISCARD_CTL_TOP +
		(area_no * OFFSET_NETTX_DISCARD_CTL_SIZE);

	return true;
}

static inline bool get_discard_rx_base_addr(u32 *base_addr, int area_no)
{
	if ((area_no < 0) || (area_no >= OFFSET_NETRX_DISCARD_CTL_MAX)) {
		return false;
	}

	*base_addr = OFFSET_NETRX_DISCARD_TOP + OFFSET_NETRX_DISCARD_CTL_TOP +
		(area_no * OFFSET_NETRX_DISCARD_CTL_SIZE);

	return true;
}

static int set_discard_enable(spinlock_t *lock, struct igw_priv *dev, u32 base_addr, u32 flags, bool enable)
{
	unsigned long irq_flags;
	const u32 addr = base_addr + OFFSET_NET_DISCARD_CTL_DISCARD_EN;
	u32 data;

	spin_lock_irqsave(lock, irq_flags);

	data = ifrtp_ior32(dev, addr);

	if (enable) {
		data |= flags;
	} else {
		data &= ~flags;
	}

	ifrtp_iow32(dev, data, addr);

	spin_unlock_irqrestore(lock, irq_flags);

	return 0;
}

static int set_discard_all_enable(struct igw_priv *dev, u32 base_addr, bool enable)
{
	const u32 addr = base_addr + OFFSET_NET_DISCARD_CTL_ALL_DISCARD_EN;

	if (enable) {
		ifrtp_iow32(dev, 0x1, addr);
	} else {
		ifrtp_iow32(dev, 0x0, addr);
	}

	return 0;
}

//----------
int ifrtp_set_tx_discard(struct igw_priv *dev, struct IocDiscardInitConfig *config)
{
	u32 base_addr;

	if (!get_discard_tx_base_addr(&base_addr, config->area_no)) {
		return -EINVAL;
	}

	set_discard_enable(&discard_tx_lock, dev, base_addr, 0xffffffff, false);
	set_discard_enable(&discard_tx_lock, dev, base_addr, config->discard_en, true);

	return 0;
}

int ifrtp_set_tx_discard_by_devindex(struct igw_priv *dev, bool enable)
{
    // const int area_no = dev->netif_priority;
	const int area_no = 0;
	u32 base_addr;
	const u32 flags =
			(0x1 << (TX_TYPE_MAIN + (dev->netif_priority & 0x1)));

	if (!get_discard_tx_base_addr(&base_addr, area_no)) {
		return -EINVAL;
	}

	set_discard_enable(&discard_tx_lock, dev, base_addr, flags, enable);

	return 0;
}

int ifrtp_set_tx_discard_all(struct igw_priv *dev, bool enable)
{
    // const int area_no = dev->netif_priority;
	const int area_no = 0;
	u32 base_addr;

	if (!get_discard_tx_base_addr(&base_addr, area_no)) {
		return -EINVAL;
	}

	set_discard_all_enable(dev, base_addr, enable);

	return 0;
}

//----------
int ifrtp_set_rx_discard(struct igw_priv *dev, struct IocDiscardInitConfig *config)
{
	u32 base_addr;

	if (!get_discard_rx_base_addr(&base_addr, config->area_no)) {
		return -EINVAL;
	}

	set_discard_enable(&discard_rx_lock, dev, base_addr, 0xffffffff, false);
	set_discard_enable(&discard_rx_lock, dev, base_addr, config->discard_en, true);

	return 0;
}

int ifrtp_set_rx_discard_by_devindex(struct igw_priv *dev, bool enable)
{
    // const int area_no = dev->netif_priority;
	const int area_no = 0;
	u32 base_addr;
	const u32 flags =
			(0x1 << (RX_TYPE_MAIN + (dev->netif_priority & 0x1)));

	if (!get_discard_rx_base_addr(&base_addr, area_no)) {
		return -EINVAL;
	}

	set_discard_enable(&discard_rx_lock, dev, base_addr, flags, enable);

	return 0;
}

int ifrtp_set_rx_discard_all(struct igw_priv *dev, bool enable)
{
    // const int area_no = dev->netif_priority;
	const int area_no = 0;
	u32 base_addr;

	if (!get_discard_rx_base_addr(&base_addr, area_no)) {
		return -EINVAL;
	}

	set_discard_all_enable(dev, base_addr, enable);

	return 0;
}

//--------------------------------------------------
// VBP
//--------------------------------------------------
int ifrtp_set_tx_vbp_default_mac_type(struct igw_priv *dev, struct IocTxVbpDefaultMacTypeInitConfig *config)
{
	unsigned int offset  = 0x0;
	unsigned int address = 0x0;

	offset = OFFSET_NETTX_VBP_TOP + OFFSET_NETTX_VBP_DEFAULT_MAC_TYPE_TOP + OFFSET_NETTX_VBP_DEFAULT_MAC_TYPE_EN;

	address = offset;
	ifrtp_iow32(dev, config->default_mac_type, address);

	return 0;
}

int ifrtp_set_tx_vbp_mac_type_bypass(struct igw_priv *dev, struct IocTxVbpMacTypeBypassInitConfig *config)
{
	int result = 0;
	unsigned int offset  = 0x0;
	unsigned int address = 0x0;

	offset = OFFSET_NETTX_VBP_TOP + OFFSET_NETTX_VBP_MAC_TYPE_BYPASS_TOP + OFFSET_NETTX_VBP_MAC_TYPE_BYPASS_EN;

	address = offset;
	ifrtp_iow32(dev, config->mac_type_bypass, address);

	return result;
}

int ifrtp_set_tx_source_secondary_mac_address(struct igw_priv *dev, struct IocNetTxSourceMacAddressConfig *config)
{
	unsigned int set_source_addr;
	unsigned int address;

#ifndef IFRTP_USE_COR
	address = OFFSET_NETTX_RTP_SUB_TOP + OFFSET_NETTX_RTP_SUB_COMMON_HEADER;
	set_source_addr = 0;
	set_source_addr |= (config->eth_src[0] << 8);
	set_source_addr |= (config->eth_src[1] << 0);
	ifrtp_iow32(dev, set_source_addr, address + OFFSET_NETTX_RTP_SUB_SRC_MAC);

	set_source_addr = 0;
	set_source_addr |= (config->eth_src[2] << 24);
	set_source_addr |= (config->eth_src[3] << 16);
	set_source_addr |= (config->eth_src[4] << 8);
	set_source_addr |= (config->eth_src[5] << 0);
	ifrtp_iow32(dev, set_source_addr, address + OFFSET_NETTX_RTP_SUB_SRC_MAC + 0x4);
#else
	address = OFFSET_NETTX_VBP_TOP + OFFSET_NETTX_VBP_COMMON_HEADER;
	set_source_addr = 0;
	set_source_addr |= (config->eth_src[0] << 8);
	set_source_addr |= (config->eth_src[1] << 0);
	ifrtp_iow32(dev, set_source_addr, address + OFFSET_NETTX_VBP_DUPLICAT_FILTER_SRC_MAC);

	set_source_addr = 0;
	set_source_addr |= (config->eth_src[2] << 24);
	set_source_addr |= (config->eth_src[3] << 16);
	set_source_addr |= (config->eth_src[4] << 8);
	set_source_addr |= (config->eth_src[5] << 0);
	ifrtp_iow32(dev, set_source_addr, address + OFFSET_NETTX_VBP_DUPLICAT_FILTER_SRC_MAC + 0x4);
#endif

	return 0;
}

int ifrtp_set_tx_source_secondary_address(struct igw_priv *dev, struct IocNetTxSourceIpAddressConfig *config)
{
	unsigned int offset_channel_top, address;

	pr_info("change ifrtp source secondary address to %pI4\n", &(config->ip_dst_addr));

#ifndef IFRTP_USE_COR
	offset_channel_top = OFFSET_NETTX_RTP_SUB_TOP + OFFSET_NETTX_RTP_SUB_COMMON_HEADER;
	address = offset_channel_top + OFFSET_NETTX_RTP_SUB_IP_SRC1;
#else
	offset_channel_top = OFFSET_NETTX_VBP_TOP + OFFSET_NETTX_VBP_DUPLICAT_FILTER_TOP + OFFSET_NETTX_VBP_COMMON_HEADER;
	address = offset_channel_top + OFFSET_NETTX_VBP_DUPLICAT_FILTER_IP_SRC1;
#endif
	ifrtp_iow32(dev, ntohl(config->ip_dst_addr), address);

	return 0;
}

int ifrtp_get_tx_source_secondary_address(struct igw_priv *dev, struct IocNetTxSourceIpAddressConfig *config)
{
	unsigned int offset_channel_top, address;

#ifndef IFRTP_USE_COR
	offset_channel_top = OFFSET_NETTX_RTP_SUB_TOP + OFFSET_NETTX_RTP_SUB_COMMON_HEADER;
	address = offset_channel_top + OFFSET_NETTX_RTP_SUB_IP_SRC1;
#else
	offset_channel_top = OFFSET_NETTX_VBP_TOP + OFFSET_NETTX_VBP_DUPLICAT_FILTER_TOP + OFFSET_NETTX_VBP_COMMON_HEADER;
	address = offset_channel_top + OFFSET_NETTX_VBP_DUPLICAT_FILTER_IP_SRC1;
#endif
	config->ip_dst_addr = htonl(ifrtp_ior32(dev, address));

	return 0;
}

#ifdef IFRTP_USE_COR
int ifrtp_set_tx_dest_secondary_mac_address(struct igw_priv *dev, struct IocNetTxDestMacAddressConfig *config)
{
	unsigned int set_dest_addr;
	unsigned int address;

	address = OFFSET_NETTX_VBP_TOP + OFFSET_NETTX_VBP_DUPLICAT_FILTER_TOP + (config->ch_no * OFFSET_NETTX_VBP_DUPLICAT_FILTER_SIZE);

	set_dest_addr = 0;
	set_dest_addr |= (config->eth_dst[0] << 24);
	set_dest_addr |= (config->eth_dst[1] << 16);
	set_dest_addr |= (config->eth_dst[2] << 8);
	set_dest_addr |= (config->eth_dst[3] << 0);
	ifrtp_iow32(dev, set_dest_addr, address + OFFSET_NETTX_VBP_DUPLICAT_FILTER_DST_MAC);

	set_dest_addr = 0;
	set_dest_addr |= (config->eth_dst[4] << 24);
	set_dest_addr |= (config->eth_dst[5] << 16);
	ifrtp_iow32(dev, set_dest_addr, address + OFFSET_NETTX_VBP_DUPLICAT_FILTER_DST_MAC + 0x4);

	return 0;
}

int ifrtp_get_tx_dest_secondary_mac_address(struct igw_priv *dev, struct IocNetTxDestMacAddressConfig *config)
{
	unsigned int dest_addr;
	unsigned int address;

	address = OFFSET_NETTX_VBP_TOP + OFFSET_NETTX_VBP_DUPLICAT_FILTER_TOP + (config->ch_no * OFFSET_NETTX_VBP_DUPLICAT_FILTER_SIZE);

	dest_addr = ifrtp_ior32(dev, address + OFFSET_NETTX_VBP_DUPLICAT_FILTER_DST_MAC);
	config->eth_dst[0] = (dest_addr >> 24) & 0xFF;
	config->eth_dst[1] = (dest_addr >> 16) & 0xFF;
	config->eth_dst[2] = (dest_addr >> 8)  & 0xFF;
	config->eth_dst[3] = (dest_addr >> 0)  & 0xFF;

	dest_addr = ifrtp_ior32(dev, address + OFFSET_NETTX_VBP_DUPLICAT_FILTER_DST_MAC + 0x4);
	config->eth_dst[4] = (dest_addr >> 24) & 0xFF;
	config->eth_dst[5] = (dest_addr >> 16) & 0xFF;

	return 0;
}

int ifrtp_set_tx_dest_secondary_address(struct igw_priv *dev, struct IocNetTxDestIpAddressConfig *config)
{
	unsigned int offset_channel_top, address;

	offset_channel_top = OFFSET_NETTX_VBP_TOP + OFFSET_NETTX_VBP_DUPLICAT_FILTER_TOP + (config->ch_no * OFFSET_NETTX_VBP_DUPLICAT_FILTER_SIZE);
	address = offset_channel_top + OFFSET_NETTX_VBP_DUPLICAT_FILTER_IP_DST1;
	ifrtp_iow32(dev, ntohl(config->ip_dst_addr), address);

	return 0;
}

int ifrtp_get_tx_dest_secondary_address(struct igw_priv *dev, struct IocNetTxDestIpAddressConfig *config)
{
	unsigned int offset_channel_top, address;

	offset_channel_top = OFFSET_NETTX_VBP_TOP + OFFSET_NETTX_VBP_DUPLICAT_FILTER_TOP + (config->ch_no * OFFSET_NETTX_VBP_DUPLICAT_FILTER_SIZE);
	address = offset_channel_top + OFFSET_NETTX_VBP_DUPLICAT_FILTER_IP_DST1;
	config->ip_dst_addr = ifrtp_ior32(dev, address);

	return 0;
}

int ifrtp_set_tx_secondary_port(struct igw_priv *dev, struct IocNetTxPortConfig *config)
{
	unsigned int offset_channel_top, address;

	offset_channel_top = OFFSET_NETTX_VBP_TOP + OFFSET_NETTX_VBP_DUPLICAT_FILTER_TOP + (config->ch_no * OFFSET_NETTX_VBP_DUPLICAT_FILTER_SIZE);
	address = offset_channel_top + OFFSET_NETTX_VBP_DUPLICAT_FILTERP_UDP_PORT;
	ifrtp_iow32(dev, config->port, address);

	return 0;
}

int ifrtp_get_tx_secondary_port(struct igw_priv *dev, struct IocNetTxPortConfig *config)
{
	unsigned int offset_channel_top, address;

	offset_channel_top = OFFSET_NETTX_VBP_TOP + OFFSET_NETTX_VBP_DUPLICAT_FILTER_TOP + (config->ch_no * OFFSET_NETTX_VBP_DUPLICAT_FILTER_SIZE);
	address = offset_channel_top + OFFSET_NETTX_VBP_DUPLICAT_FILTERP_UDP_PORT;
	config->port = ifrtp_ior32(dev, address);

	return 0;
}

int ifrtp_set_tx_duplicate_enable(struct igw_priv *dev, struct IocNetTxDuplicateEnableConfig *config)
{
	unsigned int offset_vbp_area_top, address;
	unsigned int duplicate_enable;

	offset_vbp_area_top = OFFSET_NETTX_VBP_TOP + OFFSET_NETTX_VBP_DUPLICAT_FILTER_TOP + (config->ch_no * OFFSET_NETTX_VBP_DUPLICAT_FILTER_SIZE);
	address = offset_vbp_area_top + OFFSET_NETTX_VBP_DUPLICAT_FILTER_CTL;

	if (config->enable == 1) {
#ifdef CONFIG_FPGA_APP_TYPE_NMI
		duplicate_enable = 0x0000C03F;
#else
		duplicate_enable = 0x0000803F;
#endif
	} else {
		duplicate_enable = 0x0000003F;
	}

	address = offset_vbp_area_top;
	ifrtp_iow32(dev, duplicate_enable, address);

	return 0;
}
#endif

//--------------------------------------------------
// Stream Transmitter
//--------------------------------------------------
int ifrtp_set_tx_strm_trns_chnl(struct igw_priv *dev, struct IocStrmTrnsCnlInitConfig *config)
{
	int channel = 0;
	unsigned int address = 0x0;

	channel = config->ch_no;

	if (channel < 0 || channel >= OFFSET_NETTX_STRM_TRANS_CH_MAX) {
		return -EFAULT;
	}
	address = OFFSET_NETTX_STRM_TRANS_TOP + OFFSET_NETTX_STRM_TRANS_CH_TOP + (channel * OFFSET_NETTX_STRM_TRANS_CH_SIZE);

	ifrtp_iow32(dev, config->vlan_tag, address + OFFSET_NETTX_STRM_TRANS_CH_VLAN_TAG);
	ifrtp_iow32(dev, config->ttl, address + OFFSET_NETTX_STRM_TRANS_CH_IP_HEAD);
    // TODO double check
	ifrtp_iow32(dev, config->udp_port, address + OFFSET_NETTX_STRM_TRANS_CH_UDP_PORT);

	return 0;
}

int ifrtp_set_tx_strm_trns_chnl_ctl(struct igw_priv *dev, struct IocStrmTrnsCnlCtlInitConfig *config)
{
	unsigned int offset_top = 0x0;
	unsigned int address    = 0x0;

	offset_top = OFFSET_NETTX_STRM_TRANS_TOP + OFFSET_NETTX_STRM_TRANS_CH_CTL_TOP;

	address = offset_top + OFFSET_NETTX_STRM_TRANS_CH_CTL_CH_EN;
	ifrtp_iow32(dev, config->channel_enable, address);

	address = offset_top + OFFSET_NETTX_STRM_TRANS_CH_CTL_INPLICIT_TYPE;
	ifrtp_iow32(dev, config->inplicit_type, address);

	address = offset_top + OFFSET_NETTX_STRM_TRANS_CH_CTL_IPV6_HEADER;
	ifrtp_iow32(dev, config->ipv6_header, address);

	address = offset_top + OFFSET_NETTX_STRM_TRANS_CH_CTL_IPVER_SELECT;
	ifrtp_iow32(dev, config->ipver_select, address);

	address = offset_top + OFFSET_NETTX_STRM_TRANS_CH_CTL_UNREC_CH_CNT;
	ifrtp_iow32(dev, config->unrecognized_ch_count, address);

	address = offset_top + OFFSET_NETTX_STRM_TRANS_CH_CTL_DISABLED_CH_CNT;
	ifrtp_iow32(dev, config->disabled_ch_count, address);

	address = offset_top + OFFSET_NETTX_STRM_TRANS_CH_CTL_OVERSIZED_CH_CNT;
	ifrtp_iow32(dev, config->oversized_pkt_count, address);

	return 0;
}

int ifrtp_set_tx_source_primary_mac_address(struct igw_priv *dev, struct IocNetTxSourceMacAddressConfig *config)
{
	unsigned int set_source_addr;
	unsigned int address;

	address = OFFSET_NETTX_STRM_TRANS_TOP + OFFSET_NETTX_STRM_TRANS_CH_TOP + OFFSET_NETTX_STRM_TRANS_COMMON_SRC;

	set_source_addr = 0;
	set_source_addr |= (config->eth_src[0] << 8);
	set_source_addr |= (config->eth_src[1] << 0);
	ifrtp_iow32(dev, set_source_addr, address + OFFSET_NETTX_STRM_TRANS_CH_SRC_MAC);

	set_source_addr = 0;
	set_source_addr |= (config->eth_src[2] << 24);
	set_source_addr |= (config->eth_src[3] << 16);
	set_source_addr |= (config->eth_src[4] << 8);
	set_source_addr |= (config->eth_src[5] << 0);
	ifrtp_iow32(dev, set_source_addr, address + OFFSET_NETTX_STRM_TRANS_CH_SRC_MAC + 0x4);

	return 0;
}

int ifrtp_set_tx_source_primary_address(struct igw_priv *dev, struct IocNetTxSourceIpAddressConfig *config)
{
	unsigned int offset_channel_top, address;

	pr_info("change ifrtp source primary address to %pI4\n", &(config->ip_dst_addr));
	offset_channel_top = OFFSET_NETTX_STRM_TRANS_TOP + OFFSET_NETTX_STRM_TRANS_CH_TOP + OFFSET_NETTX_STRM_TRANS_COMMON_SRC;
	address = offset_channel_top + OFFSET_NETTX_STRM_TRANS_CH_IP_SRC1;
	ifrtp_iow32(dev, ntohl(config->ip_dst_addr), address);

	return 0;
}

int ifrtp_get_tx_source_primary_address(struct igw_priv *dev, struct IocNetTxSourceIpAddressConfig *config)
{
	unsigned int offset_channel_top, address;

	offset_channel_top = OFFSET_NETTX_STRM_TRANS_TOP + OFFSET_NETTX_STRM_TRANS_CH_TOP + OFFSET_NETTX_STRM_TRANS_COMMON_SRC;
	address = offset_channel_top + OFFSET_NETTX_STRM_TRANS_CH_IP_SRC1;
	config->ip_dst_addr = htonl(ifrtp_ior32(dev, address));

	return 0;
}

int ifrtp_set_tx_dest_primary_mac_address(struct igw_priv *dev, struct IocNetTxDestMacAddressConfig *config)
{
	unsigned int set_dest_addr;
	unsigned int address;

	address = OFFSET_NETTX_STRM_TRANS_TOP + OFFSET_NETTX_STRM_TRANS_CH_TOP + (config->ch_no * OFFSET_NETTX_STRM_TRANS_CH_SIZE);

	set_dest_addr = 0;
	set_dest_addr |= (config->eth_dst[0] << 24);
	set_dest_addr |= (config->eth_dst[1] << 16);
	set_dest_addr |= (config->eth_dst[2] << 8);
	set_dest_addr |= (config->eth_dst[3] << 0);
	ifrtp_iow32(dev, set_dest_addr, address + OFFSET_NETTX_STRM_TRANS_CH_DST_MAC);

	set_dest_addr = 0;
	set_dest_addr |= (config->eth_dst[4] << 24);
	set_dest_addr |= (config->eth_dst[5] << 16);
	ifrtp_iow32(dev, set_dest_addr, address + OFFSET_NETTX_STRM_TRANS_CH_DST_MAC + 0x4);

	return 0;
}

int ifrtp_get_tx_dest_primary_mac_address(struct igw_priv *dev, struct IocNetTxDestMacAddressConfig *config)
{
	unsigned int dest_addr;
	unsigned int address;

	address = OFFSET_NETTX_STRM_TRANS_TOP + OFFSET_NETTX_STRM_TRANS_CH_TOP + (config->ch_no * OFFSET_NETTX_STRM_TRANS_CH_SIZE);

	dest_addr = ifrtp_ior32(dev, address + OFFSET_NETTX_STRM_TRANS_CH_DST_MAC);
	config->eth_dst[0] = (dest_addr >> 24) & 0xFF;
	config->eth_dst[1] = (dest_addr >> 16) & 0xFF;
	config->eth_dst[2] = (dest_addr >> 8)  & 0xFF;
	config->eth_dst[3] = (dest_addr >> 0)  & 0xFF;

	dest_addr = ifrtp_ior32(dev, address + OFFSET_NETTX_STRM_TRANS_CH_DST_MAC + 0x4);
	config->eth_dst[4] = (dest_addr >> 24) & 0xFF;
	config->eth_dst[5] = (dest_addr >> 16) & 0xFF;

	return 0;
}

int ifrtp_set_tx_dest_primary_address(struct igw_priv *dev, struct IocNetTxDestIpAddressConfig *config)
{
	unsigned int offset_channel_top, address;

	offset_channel_top = OFFSET_NETTX_STRM_TRANS_TOP + OFFSET_NETTX_STRM_TRANS_CH_TOP + (config->ch_no * OFFSET_NETTX_STRM_TRANS_CH_SIZE);
	address = offset_channel_top + OFFSET_NETTX_STRM_TRANS_CH_IP_DST1;
	ifrtp_iow32(dev, ntohl(config->ip_dst_addr), address);

	return 0;
}

int ifrtp_get_tx_dest_primary_address(struct igw_priv *dev, struct IocNetTxDestIpAddressConfig *config)
{
	unsigned int offset_channel_top, address;

	offset_channel_top = OFFSET_NETTX_STRM_TRANS_TOP + OFFSET_NETTX_STRM_TRANS_CH_TOP + (config->ch_no * OFFSET_NETTX_STRM_TRANS_CH_SIZE);
	address = offset_channel_top + OFFSET_NETTX_STRM_TRANS_CH_IP_DST1;
	config->ip_dst_addr = ifrtp_ior32(dev, address);

	return 0;
}

int ifrtp_set_tx_primary_port(struct igw_priv *dev, struct IocNetTxPortConfig *config)
{
	unsigned int offset_channel_top, address;

	offset_channel_top = OFFSET_NETTX_STRM_TRANS_TOP + OFFSET_NETTX_STRM_TRANS_CH_TOP + (config->ch_no * OFFSET_NETTX_STRM_TRANS_CH_SIZE);
	address = offset_channel_top + OFFSET_NETTX_STRM_TRANS_CH_UDP_PORT;
	ifrtp_iow32(dev, config->port, address);

	return 0;
}

int ifrtp_get_tx_primary_port(struct igw_priv *dev, struct IocNetTxPortConfig *config)
{
	unsigned int offset_channel_top, address;

	offset_channel_top = OFFSET_NETTX_STRM_TRANS_TOP + OFFSET_NETTX_STRM_TRANS_CH_TOP + (config->ch_no * OFFSET_NETTX_STRM_TRANS_CH_SIZE);
	address = offset_channel_top + OFFSET_NETTX_STRM_TRANS_CH_UDP_PORT;
	config->port = ifrtp_ior32(dev, address);

	return 0;
}

//--------------------------------------------------
// Validation
//--------------------------------------------------
int ifrtp_set_rx_validation(struct igw_priv *dev, struct IocPktValidationRxInitConfig *config)
{
	int result  = 0;
	int area_no = 0;
	unsigned int offset_area_top = 0x0;
	unsigned int address         = 0x0;

	area_no = config->area_no;

	if (area_no >= 0 && area_no < OFFSET_NETRX_VALIDATE_CTL_MAX) {
		offset_area_top = OFFSET_NETRX_VALIDATE_TOP + OFFSET_NETRX_VALIDATE_CTL_TOP + (area_no * OFFSET_NETRX_VALIDATE_CTL_SIZE);

		address = offset_area_top + OFFSET_NETRX_VALIDATE_CTL_PKT_DISCARD_FLG;
		ifrtp_iow32(dev, config->chksum_err_discard_en, address);

		address = offset_area_top + OFFSET_NETRX_VALIDATE_CTL_PKT_LEN_ERR_DISCARD_FLG;
		ifrtp_iow32(dev, config->pktlen_err_discard_en, address);
	} else {
		result = -EFAULT;
	}

	return result;
}

//--------------------------------------------------
// Extraction
//--------------------------------------------------
int ifrtp_set_rx_extraction(struct igw_priv *dev, struct IocPayloadExtractionRxInitConfig *config)
{
	int result  = 0;
	int area_no = 0;
	unsigned int offset_area_top = 0x0;
	unsigned int address         = 0x0;

	area_no = config->area_no;

	if (area_no >= 0 && area_no < OFFSET_NETRX_PAYLOAD_EXTRACT_CTL_MAX) {
		offset_area_top = OFFSET_NETRX_PAYLOAD_EXTRACT_TOP + OFFSET_NETRX_PAYLOAD_EXTRACT_CTL_TOP + (area_no * OFFSET_NETRX_PAYLOAD_EXTRACT_CTL_SIZE);
		address = offset_area_top + OFFSET_NETRX_PAYLOAD_EXTRACT_CTL_EXTRACT_FLG_EN;
		ifrtp_iow32(dev, config->all_type_en, address);

		if (config->enable_default_type) {
			address = offset_area_top + OFFSET_NETRX_PAYLOAD_EXTRACT_CTL_EXTRACT_DEFAULT_MEDIA_TYPE;
			ifrtp_iow32(dev, config->default_type, address);
		}

		address = offset_area_top + OFFSET_NETRX_PAYLOAD_EXTRACT_CTL_EXTRACT_DEFAULT_PASS;
		ifrtp_iow32(dev, config->enable_default_type, address);

#ifdef CONFIG_FPGA_APP_TYPE_NMI
		address = offset_area_top + OFFSET_NETRX_PAYLOAD_EXTRACT_CTL_EXTRACT_NMI_MODE_EN;
		ifrtp_iow32(dev, 0x00ffffff, address);
#endif
	} else {
		result = -EFAULT;
	}

	return result;
}

void ifrtp_set_rx_extraction_dual_ssrc(struct igw_priv *dev, struct IocPayloadExtractionRxSSRCConfig *config)
{
	unsigned int read_reg;
	unsigned int offset_area_top = 0x0;
	unsigned int address         = 0x0;

	offset_area_top = OFFSET_NETRX_PAYLOAD_EXTRACT_TOP + OFFSET_NETRX_PAYLOAD_EXTRACT_CTL_TOP + (config->area_no * OFFSET_NETRX_PAYLOAD_EXTRACT_CTL_SIZE);
	address = offset_area_top + OFFSET_NETRX_PAYLOAD_EXTRACT_CTL_EXTRACT_DUAL_TYPE;
	read_reg = ifrtp_ior32(dev, address);

	if (config->dual_en) {
		ifrtp_iow32(dev, read_reg | 1 << config->nwk_pkt_type, address);

		// set ssrc
		address = offset_area_top + OFFSET_NETRX_PAYLOAD_EXTRACT_CTL_EXTRACT_MEDIA_TYPE;
		ifrtp_iow32(dev, config->media_type, address);

		address = offset_area_top + OFFSET_NETRX_PAYLOAD_EXTRACT_CTL_EXTRACT_PACKET_TYPE;
		ifrtp_iow32(dev, config->nwk_pkt_type, address);

		address = offset_area_top + OFFSET_NETRX_PAYLOAD_EXTRACT_CTL_EXTRACT_SYNC_SOURCE;
		ifrtp_iow32(dev, config->ssrc, address);
	} else {
		ifrtp_iow32(dev, read_reg & ~(1 << config->nwk_pkt_type), address);
	}
}

int ifrtp_get_rx_sequence_number_error_status(struct igw_priv *dev, struct IocNetRxSequenceNumberErrorStatusConfig *config)
{
	unsigned int offset, address;
	int result = 0;

	offset = OFFSET_NETRX_PAYLOAD_EXTRACT_TOP + OFFSET_NETRX_PAYLOAD_EXTRACT_CTL_TOP + OFFSET_NETRX_PAYLOAD_EXTRACT_CTL_SIZE * config->engine_no;
	address = offset + OFFSET_NETRX_PAYLOAD_EXTRACT_CTL_PKT_LOSS_MONITOR;
	config->sequence_number_error = ifrtp_ior32(dev, address) & config->sequence_number_error_mask;
	ifrtp_iow32(dev, config->sequence_number_error, address);
	return result;
}

//--------------------------------------------------
// Identification
//--------------------------------------------------
int ifrtp_set_rx_identification(struct igw_priv *dev, struct IocIdentificationRxInitConfig *config)
{
	int result = 0;
	int area_no    = 0;
	int ch_type_no = 0;
	unsigned int offset_area_top    = 0x0;
	unsigned int offset_ch_type_top = 0x0;
	unsigned int address            = 0x0;

	area_no = config->area_no;

	if (area_no >= 0 && area_no < OFFSET_NETRX_IDENT_CTL_MAX) {
		offset_area_top = OFFSET_NETRX_IDENT_TOP + OFFSET_NETRX_IDENT_CTL_TOP + (area_no * OFFSET_NETRX_IDENT_CTL_SIZE);

		address = offset_area_top + OFFSET_NETRX_IDENT_CTL_FILTER_EN;
		ifrtp_iow32(dev, config->pkt_filter_en, address);
		address = offset_area_top + OFFSET_NETRX_IDENT_CTL_SRC_IP_FILTER_EN;
		ifrtp_iow32(dev, config->src_ip_addr_en, address);

		address = offset_area_top + OFFSET_NETRX_IDENT_CTL_DST_IP_FILTER_EN;
		ifrtp_iow32(dev, config->dst_ip_addr_en, address);

		address = offset_area_top + OFFSET_NETRX_IDENT_CTL_DST_PORT_FILTER_EN;
		ifrtp_iow32(dev, config->dst_port_en, address);

		address = offset_area_top + OFFSET_NETRX_IDENT_CTL_PKT_FILTER_TYPE_DEF;
		ifrtp_iow32(dev, config->unmatched_type, address);

		offset_ch_type_top = offset_area_top + OFFSET_NETRX_IDENT_CTL_PKT_FILTER_TYPE_TOP;
		for (ch_type_no = 0; ch_type_no < OFFSET_NETRX_IDENT_CTL_PKT_FILTER_TYPE_MAX; ch_type_no++) {
			address = offset_ch_type_top + (ch_type_no * OFFSET_NETRX_IDENT_CTL_PKT_FILTER_TYPE_SIZE);
			ifrtp_iow32(dev, config->ch_type[ch_type_no], address);
		}
	}

	return result;
}

int ifrtp_get_rx_identification(struct igw_priv *dev, struct IocIdentificationRxInitConfig *config)
{
	int result     = 0;
	int area_no    = 0;
	int ch_type_no = 0;
	unsigned int offset_area_top    = 0x0;
	unsigned int offset_ch_type_top = 0x0;
	unsigned int address            = 0x0;

	area_no = config->area_no;

	if (area_no >= 0 && area_no < OFFSET_NETRX_IDENT_CTL_MAX) {
		offset_area_top = OFFSET_NETRX_IDENT_TOP + OFFSET_NETRX_IDENT_CTL_TOP + (area_no * OFFSET_NETRX_IDENT_CTL_SIZE);

		address = offset_area_top + OFFSET_NETRX_IDENT_CTL_FILTER_EN;
		config->pkt_filter_en = ifrtp_ior32(dev, address);
		address = offset_area_top + OFFSET_NETRX_IDENT_CTL_SRC_IP_FILTER_EN;
		config->src_ip_addr_en = ifrtp_ior32(dev, address);

		address = offset_area_top + OFFSET_NETRX_IDENT_CTL_DST_IP_FILTER_EN;
		config->dst_ip_addr_en = ifrtp_ior32(dev, address);

		address = offset_area_top + OFFSET_NETRX_IDENT_CTL_DST_PORT_FILTER_EN;
		config->dst_port_en = ifrtp_ior32(dev, address);

		address = offset_area_top + OFFSET_NETRX_IDENT_CTL_PKT_FILTER_TYPE_DEF;
		config->unmatched_type = ifrtp_ior32(dev, address);

		offset_ch_type_top = offset_area_top + OFFSET_NETRX_IDENT_CTL_PKT_FILTER_TYPE_TOP;
		for (ch_type_no = 0; ch_type_no < OFFSET_NETRX_IDENT_CTL_PKT_FILTER_TYPE_MAX; ch_type_no++) {
			address = offset_ch_type_top + (ch_type_no * OFFSET_NETRX_IDENT_CTL_PKT_FILTER_TYPE_SIZE);
			config->ch_type[ch_type_no] = ifrtp_ior32(dev, address);
		}
	}

	return result;
}

int ifrtp_set_rx_dest_address(struct igw_priv *dev, struct IocNetRxDestIpAddressConfig *config)
{
	unsigned int offset_area_top, address;

	offset_area_top = OFFSET_NETRX_IDENT_TOP + OFFSET_NETRX_IDENT_CTL_TOP + OFFSET_NETRX_IDENT_CTL_ADDR_TOP +
		(config->module_no * OFFSET_NETRX_IDENT_CTL_SIZE) + (config->filter_no * OFFSET_NETRX_IDENT_CTL_ADDR_SIZE);
	address = offset_area_top + OFFSET_NETRX_IDENT_CTL_ADDR_DST_IP;
	ifrtp_iow32(dev, ntohl(config->ip_dst_addr), address);

	return 0;
}

int ifrtp_get_rx_dest_address(struct igw_priv *dev, struct IocNetRxDestIpAddressConfig *config)
{
	unsigned int offset_area_top, address;

	offset_area_top = OFFSET_NETRX_IDENT_TOP + OFFSET_NETRX_IDENT_CTL_TOP + OFFSET_NETRX_IDENT_CTL_ADDR_TOP +
		(config->module_no * OFFSET_NETRX_IDENT_CTL_SIZE) + (config->filter_no * OFFSET_NETRX_IDENT_CTL_ADDR_SIZE);
	address = offset_area_top + OFFSET_NETRX_IDENT_CTL_ADDR_DST_IP;
	config->ip_dst_addr = ifrtp_ior32(dev, address);

	return 0;
}

int ifrtp_set_rx_port(struct igw_priv *dev, struct IocNetRxPortConfig *config)
{
	unsigned int offset_area_top, address;

	offset_area_top = OFFSET_NETRX_IDENT_TOP + OFFSET_NETRX_IDENT_CTL_TOP + OFFSET_NETRX_IDENT_CTL_ADDR_TOP +
		(config->module_no * OFFSET_NETRX_IDENT_CTL_SIZE) + (config->filter_no * OFFSET_NETRX_IDENT_CTL_ADDR_SIZE);
	address = offset_area_top + OFFSET_NETRX_IDENT_CTL_ADDR_DST_PORT;

	ifrtp_iow32(dev, config->port, address);

	return 0;
}

int ifrtp_get_rx_port(struct igw_priv *dev, struct IocNetRxPortConfig *config)
{
	unsigned int offset_area_top, address;

	offset_area_top = OFFSET_NETRX_IDENT_TOP + OFFSET_NETRX_IDENT_CTL_TOP + OFFSET_NETRX_IDENT_CTL_ADDR_TOP +
		(config->module_no * OFFSET_NETRX_IDENT_CTL_SIZE) + (config->filter_no * OFFSET_NETRX_IDENT_CTL_ADDR_SIZE);
	address = offset_area_top + OFFSET_NETRX_IDENT_CTL_ADDR_DST_PORT;

	config->port = ifrtp_ior32(dev, address);

	return 0;
}

int ifrtp_set_rx_filter_enable(struct igw_priv *dev)
{
	unsigned int offset_area_top, address;
	unsigned int set_data = 0xC3FFFFFF;

	offset_area_top = OFFSET_NETRX_IDENT_TOP + OFFSET_NETRX_IDENT_CTL_TOP + OFFSET_NETRX_IDENT_CTL_FILTER_EN;
	address = offset_area_top;

	ifrtp_iow32(dev, set_data, address);
	return 0;
}

int ifrtp_enable_stream_filter(struct igw_priv *dev)
{
    // Drop all unmatched UDP packets
	static const u32 FIREWALL_FILTER_DST_ADDR_LOW;
	static const u32 FIREWALL_FILTER_DST_ADDR_HIGH = 0xFFFFFFFF;
	unsigned int offset;
	int i;

	for (i = 0; i < IFRTP_PHY_COUNT; i++) {
		offset = OFFSET_NETRX_IDENT_TOP + OFFSET_NETRX_IDENT_CTL_TOP + OFFSET_NETRX_IDENT_CTL_SIZE * i;
		ifrtp_iow32(dev, FIREWALL_FILTER_DST_ADDR_LOW,  offset + OFFSET_NETRX_IDENT_CTL_FIREWALL_FILTER_DST_ADDR_LOW);
		ifrtp_iow32(dev, FIREWALL_FILTER_DST_ADDR_HIGH, offset + OFFSET_NETRX_IDENT_CTL_FIREWALL_FILTER_DST_ADDR_HIGH);
		ifrtp_iow32(dev, RX_TYPE_FIREWALL_FILTER, offset + OFFSET_NETRX_IDENT_CTL_FIREWALL_FILTER_PKT_TYPE);
		ifrtp_iow32(dev, 0x1, offset + OFFSET_NETRX_IDENT_CTL_FIREWALL_FILTER_EN);
	}

	return 0;
}

//--------------------------------------------------
// TRM
//--------------------------------------------------

int ifrtp_enable_rx_trm_bucket(struct igw_priv *dev)
{
	unsigned int address = 0x0;

	address = OFFSET_TRM_RX_TOP;

	ifrtp_iow32(dev, 0xffffffff, address);
	ifrtp_iow32(dev, 0xffffffff, address + 0x4);
	ifrtp_iow32(dev, 0x00000000, address + 0x8);

	return 0;
}

int ifrtp_init_rx_trm_ctrl_bucket_init(struct igw_priv *dev)
{
	unsigned int address;

	address = OFFSET_TRM_RX_TOP + 0x4;
	ifrtp_iow32(dev, 0xffffffff, address);

	return 0;

}

int ifrtp_set_rx_trm_ctrl_bucket_init(struct igw_priv *dev, struct IocNetRxTrmCtrlBucketInitConfig *config)
{
	unsigned int address;

	address = OFFSET_TRM_RX_TOP + 0x4;
	ifrtp_iow32(dev, config->nwk_pkt_type, address);

	return 0;
}

unsigned int ifrtp_get_rx_xgmac_good_frame_lower(struct igw_priv *dev)
{
	const unsigned int address = OFFSET_NETRX_XGMAC_TOP + OFFSET_NETRX_XGMAC_CTL_TOP +
		(dev->netif_priority * OFFSET_NETRX_XGMAC_CTL_SIZE) + OFFSET_NETRX_XGMAC_CTL_RECV_GOOD_FRM_LOWER_CNT;
	return ifrtp_ior32(dev, address);
}

void ifrtp_transmitter_system_enable(struct igw_priv *dev)
{
	ifrtp_iow32(dev, 0x1, OFFSET_NETTX_TRANS_SYS_TOP + OFFSET_NETTX_TRANS_SYS_EN);
}

void ifrtp_receiver_system_enable(struct igw_priv *dev)
{
	ifrtp_iow32(dev, 0x3, OFFSET_RECV_SYSTEM_TOP);
}

