// SPDX-License-Identifier: GPL-2.0
/*
 * Sony ifrtp SFP Ethernet device driver
 *
 * Copyright (c) 2023 Sony Corporation. All rights reserved.
 */
#include "ifrtp.h"
#include "igw_driver.h"

#include "io.h"
#include "igw_net_address.h"
#include "igw_netreg.h"

extern struct ifrtp_config ifrtp_cfg;

static void init_tx_strmtrans_ch_setting(struct igw_priv *dev)
{
	struct IocStrmTrnsCnlInitConfig api_config;
	int i;

	api_config.vlan_tag       = 0x70110000;
	api_config.ttl            = 64;
	api_config.ipv6_flow      = 0x000003E8;
	api_config.ip_src_addr[0] = 0;
	api_config.ip_src_addr[1] = 0;
	api_config.ip_src_addr[2] = 0;
	api_config.ip_src_addr[3] = 0;
	api_config.ip_dst_addr[0] = 0;
	api_config.ip_dst_addr[1] = 0;
	api_config.ip_dst_addr[2] = 0;
	api_config.ip_dst_addr[3] = 0;
	api_config.udp_port       = 0;

	for (i = 0; i < NN_NWK_PKT_TYPE_MAX; i++) {
		api_config.ch_no = i;
		ifrtp_set_tx_strm_trns_chnl(dev, &api_config);
	}
}

static void init_tx_strmtrans_ch_ctl(struct igw_priv *dev)
{
	struct IocStrmTrnsCnlCtlInitConfig api_config;

	api_config.channel_enable        = 0x00FFFFFF;
	api_config.inplicit_type         = 0x00000000;
	api_config.ipv6_header           = 0x00000000;
	api_config.ipver_select          = 0x00000000;
	api_config.unrecognized_ch_count = 0x00000000;
	api_config.disabled_ch_count     = 0x00000000;
	api_config.oversized_pkt_count   = 0x00000000;

	ifrtp_set_tx_strm_trns_chnl_ctl(dev, &api_config);
}

static void init_tx_packet_unpack(struct igw_priv *dev)
{
	struct IocPktUnpackTxDirInitConfig api_config;

	api_config.unpack_enable = 0xFC000000;

	ifrtp_set_tx_pkt_unpack(dev, &api_config);
}

static void init_tx_vbp_default_mac_type(struct igw_priv *dev)
{
	struct IocTxVbpDefaultMacTypeInitConfig api_config;

	api_config.default_mac_type = 0x80000000;

	ifrtp_set_tx_vbp_default_mac_type(dev, &api_config);
}

static void init_tx_vbp_mac_type_bypass(struct igw_priv *dev)
{
	struct IocTxVbpMacTypeBypassInitConfig api_config;

	api_config.mac_type_bypass = 0x03000000;

	ifrtp_set_tx_vbp_mac_type_bypass(dev, &api_config);
}

static void init_tx_packet_pack(struct igw_priv *dev)
{
	struct IocPktPackTxDirInitConfig api_config;

	api_config.packet_pack_no = 0;
	api_config.pack_enable    = 0xFFFFFFFF;
	ifrtp_set_tx_pkt_pack(dev, &api_config);

#ifdef IFRTP_USE_COR
	api_config.packet_pack_no = 1;
	ifrtp_set_tx_pkt_pack(dev, &api_config);
#endif
}

static int ifrtp_set_tx_dest_primary_mac_address_u64(struct igw_priv *dev, unsigned int ch_no, u64 eth_dst)
{
	u32 set_dest_addr;
	unsigned int address;

	address = OFFSET_NETTX_STRM_TRANS_TOP + OFFSET_NETTX_STRM_TRANS_CH_TOP + (ch_no * OFFSET_NETTX_STRM_TRANS_CH_SIZE);

	set_dest_addr = eth_dst >> 32;
	ifrtp_iow32(dev, set_dest_addr, address + OFFSET_NETTX_STRM_TRANS_CH_DST_MAC);

	set_dest_addr = eth_dst;
	ifrtp_iow32(dev, set_dest_addr, address + OFFSET_NETTX_STRM_TRANS_CH_DST_MAC + 0x4);

	return 0;
}

static void init_tx_dest_primary_address(struct igw_priv *dev)
{
	struct IocNetTxDestIpAddressConfig ip_config = {.ip_dst_addr = ifrtp_cfg.target_ip};
	int i;
	u64 eth_dst = 0;

	for (i = 0; i < OFFSET_NETTX_STRM_TRANS_CH_MAX; ++i) {
		ifrtp_set_tx_dest_primary_mac_address_u64(dev, i, eth_dst);

		ip_config.ch_no = i;
		ifrtp_set_tx_dest_primary_address(dev, &ip_config);
	}
}

#ifndef IFRTP_USE_COR
static void init_tx_mac_selector(struct igw_priv *dev)
{
	unsigned int address;

	address = OFFSET_NETTX_RTP_SUB_TOP + OFFSET_NETTX_RTP_SUB_COMMON_HEADER;
	ifrtp_iow32(dev, ifrtp_cfg.main_mac_mask, address + OFFSET_NETTX_RTP_SUB_FILTER_MAC0);
	ifrtp_iow32(dev, 0xF0FFFFFF, address + OFFSET_NETTX_RTP_SUB_FILTER_MAC1);
}
#endif

static void init_rx_trm_vsw_init(struct igw_priv *dev)
{
	ifrtp_enable_rx_trm_bucket(dev);
	ifrtp_init_rx_trm_ctrl_bucket_init(dev);
}

static void init_rx_pkt_unpack(struct igw_priv *dev)
{
	struct IocPktUnpackRxInitConfig api_config;

	api_config.area_no                      = 0;
	api_config.unrecognized_next_header_cnt = 0x00000000;
	api_config.unrecognized_frame_type_cnt  = 0x00000000;
	api_config.packet_size_mismatch_cnt     = 0x00000000;
	api_config.dumped_packet_cnt            = 0x00000000;
	api_config.unpack_every_type            = 0xFFFFFFFF;
	ifrtp_set_rx_pkt_unpack(dev, &api_config);

#ifdef IFRTP_USE_COR
	api_config.area_no                      = 1;
	ifrtp_set_rx_pkt_unpack(dev, &api_config);
#endif
}

static void init_rx_identification(struct igw_priv *dev)
{
	struct IocIdentificationRxInitConfig api_config;
	struct IocNetRxPortConfig port_config;
	struct IocNetRxDestIpAddressConfig multicast_filter_config;
	int i, j;

	api_config.page           = 0;
	api_config.pkt_filter_en  = 0xFFFFFFFF;
	api_config.src_ip_addr_en = 0xFFFFFFFF;

    // Bypass Destination IP Address Filter for non-Multicast Addr
	api_config.dst_ip_addr_en = 0xFCFFFFFF;

    // Bypass Destination Port Filter for Multicast Addr
	api_config.dst_port_en    = 0x03000000;

    // drop unknown packet
    // api_config.unmatched_type = 0x0000001D;

	for (i = 0; i < IFRTP_SDI_CH_MAX_PER_CORE; ++i) {
		api_config.ch_type[i] = i;
	}

	api_config.ch_type[26]    = 0x00000018;
	api_config.ch_type[27]    = 0x00000018;
    // api_config.ch_type[28]    = 0x00000020;
    // api_config.ch_type[29]    = 0x00000020;
	api_config.ch_type[28]    = 0x0000001D;
	api_config.ch_type[29]    = 0x0000001D;

	for (i = 0; i < IFRTP_PHY_COUNT; i++) {
		api_config.area_no        = i;
		api_config.unmatched_type = RX_TYPE_MAIN + i;

		for (j = IFRTP_SDI_CH_MAX_PER_CORE; j < 26; ++j) {
			api_config.ch_type[j] = RX_TYPE_MAIN + i;
		}
		api_config.ch_type[30]    = RX_TYPE_MAIN + i;
		api_config.ch_type[31]    = RX_TYPE_MAIN + i;

		ifrtp_set_rx_identification(dev, &api_config);

		// MulticastAddr Filter
		multicast_filter_config.module_no = i;
		multicast_filter_config.filter_no = 24;
		multicast_filter_config.ip_dst_addr = htonl(0xE0000001); //224.0.0.1
		ifrtp_set_rx_dest_address(dev, &multicast_filter_config);

		// mDNS
		multicast_filter_config.filter_no = 25;
		multicast_filter_config.ip_dst_addr = htonl(0xE00000FB); //224.0.0.251
		ifrtp_set_rx_dest_address(dev, &multicast_filter_config);

		port_config.module_no = i;
		port_config.filter_no = 26;
		port_config.port = 0x13F;
		ifrtp_set_rx_port(dev, &port_config);
		port_config.filter_no = 27;
		port_config.port = 0x140;
		ifrtp_set_rx_port(dev, &port_config);
	}
}

static void init_rx_dest_port(struct igw_priv *dev)
{
	struct IocNetRxPortConfig port_config;
	int i;

	port_config.module_no = 0;
	port_config.port = 0;

	for (i = 0; i < OFFSET_NETTX_STRM_TRANS_CH_MAX; ++i) {
		port_config.filter_no = i;
		ifrtp_set_rx_port(dev, &port_config);
	}
}

static void init_rx_validation(struct igw_priv *dev)
{
	struct IocPktValidationRxInitConfig api_config;

	api_config.area_no               = 0;
	api_config.page                  = 0;
	api_config.ip_header_chksum_en   = 0xFFFFFFFF;
	api_config.tcp_chksum_en         = 0xFFFFFFFF;
	api_config.udp_chksum_en         = 0xFFFFFFFF;
	api_config.chksum_err_discard_en = 0xFFFFFFFF;
	api_config.pktlen_err_discard_en = 0x01FFFFFF;
	ifrtp_set_rx_validation(dev, &api_config);

	api_config.area_no = 1;
	ifrtp_set_rx_validation(dev, &api_config);
}

static void init_rx_pkt_pack(struct igw_priv *dev)
{
	struct IocPktPackRxInitConfig api_config;

	api_config.area_no                      = 0;
	api_config.unrecognized_next_header_cnt = 0x00000000;
	api_config.unrecognized_frame_type_cnt  = 0x00000000;
	api_config.packet_size_mismatch_cnt     = 0x00000000;
	api_config.dumped_packet_cnt            = 0x00000000;
	api_config.no_packing                   = 0xFF000000;
	ifrtp_set_rx_pkt_pack(dev, &api_config);

    // api_config.area_no = 1;
    // ifrtp_set_rx_pkt_pack(dev, &api_config);

}

static void init_rx_extraction(struct igw_priv *dev)
{
	struct IocPayloadExtractionRxInitConfig api_config;

	api_config.area_no     = 0;
	api_config.page        = 0;
    // api_config.header_en   = 0x00FFFFFF;
    // api_config.preview_en  = 0x00FFFFFF;
	api_config.default_type = 29;
	api_config.enable_default_type = 1;
#ifdef IFRTP_USE_COR
	api_config.all_type_en = 0x00FFFFFF;
#else
	api_config.all_type_en = 0xFF000000;
#endif
	ifrtp_set_rx_extraction(dev, &api_config);

#ifdef IFRTP_USE_COR
	api_config.area_no = 1;
	ifrtp_set_rx_extraction(dev, &api_config);
#endif
}

static void init_rx_identification_addr(struct igw_priv *dev)
{
	int i;
	struct IocNetRxDestIpAddressConfig ip_config = {
		.module_no = 0,
		.ip_dst_addr = ifrtp_cfg.target_ip
	};

	for (i = 0; i < OFFSET_NETTX_STRM_TRANS_CH_MAX; ++i) {
		ip_config.filter_no = i;
		ifrtp_set_rx_dest_address(dev, &ip_config);
	}
}

void ifrtp_init_network_device_tx(struct igw_priv *dev)
{
    // Stream Transmitter Module
	init_tx_strmtrans_ch_setting(dev);

    // +-- channel controls
	init_tx_strmtrans_ch_ctl(dev);

    // Packet Unpack Module #0
	init_tx_packet_unpack(dev);

#ifndef IFRTP_USE_COR
	init_tx_mac_selector(dev);
#else
    // +-- Defualt MAC Type
	init_tx_vbp_default_mac_type(dev);

    // +-- Bypass MAC Type
	init_tx_vbp_mac_type_bypass(dev);
#endif

    // Packet Pack Module
    // +-- Tx direction
	init_tx_packet_pack(dev);

	init_tx_dest_primary_address(dev);
}

void ifrtp_init_network_device_rx(struct igw_priv *dev)
{
#ifdef IFRTP_USE_COR
	init_rx_trm_vsw_init(dev);
#endif

	init_rx_pkt_unpack(dev);

	init_rx_identification(dev);

	init_rx_dest_port(dev);

    // init_rx_validation(dev);

	init_rx_pkt_pack(dev);

	init_rx_extraction(dev);

	init_rx_identification_addr(dev);
}
