// SPDX-License-Identifier: GPL-2.0
/*
 * Sony ifrtp SFP Ethernet device driver
 *
 * Copyright (c) 2023 Sony Corporation. All rights reserved.
 */
#include "io.h"
#include "igw_net_address.h"

#define MASKBIT_IRQ_L1_NET             (0x000000001 << 0)

#define MASKBIT_IRQ_L2_NET_RX          (0x000000001 << 0)
#define MASKBIT_IRQ_L2_NET_RAW_PACKET  (0x000000001 << 1)

static DEFINE_SPINLOCK(ifrtp_irq_lock);

static void ifrtp_enable_irq(struct igw_priv *dev,  unsigned int offset, unsigned int irq_mask)
{
	unsigned long flags;
	unsigned int mask;

	spin_lock_irqsave(&ifrtp_irq_lock, flags);
	mask = ifrtp_ior32(dev, offset) | irq_mask;
	ifrtp_iow32(dev, mask, offset);
	spin_unlock_irqrestore(&ifrtp_irq_lock, flags);
}

void ifrtp_enable_net_raw_packet_irq(struct igw_priv *dev)
{
	ifrtp_enable_irq(dev, OFFSET_IRQ_L2_NET_ENABLE, MASKBIT_IRQ_L2_NET_RAW_PACKET);
}

void ifrtp_enable_net_rx_irq(void)
{
	ifrtp_enable_irq(NULL, OFFSET_IRQ_L2_NET_ENABLE, MASKBIT_IRQ_L2_NET_RX);
}

static void ifrtp_disable_irq(struct igw_priv *dev,  unsigned int offset, unsigned int irq_mask)
{
	unsigned long flags;
	unsigned int mask;

	spin_lock_irqsave(&ifrtp_irq_lock, flags);
	mask = ifrtp_ior32(dev, offset) & ~irq_mask;
	ifrtp_iow32(dev, mask, offset);
	spin_unlock_irqrestore(&ifrtp_irq_lock, flags);
}

void ifrtp_disable_all_net_irq(struct igw_priv *dev)
{
    // only disable L2 for now
	ifrtp_iow32(dev, 0, OFFSET_IRQ_L2_NET_ENABLE);
}


void ifrtp_disable_net_raw_packet_irq(struct igw_priv *dev)
{
	ifrtp_disable_irq(dev, OFFSET_IRQ_L2_NET_ENABLE, MASKBIT_IRQ_L2_NET_RAW_PACKET);
}

void ifrtp_disable_net_rx_irq(void)
{
	ifrtp_disable_irq(NULL, OFFSET_IRQ_L2_NET_ENABLE, MASKBIT_IRQ_L2_NET_RX);
}

static bool ifrtp_get_irq_status(struct igw_priv *dev, unsigned int offset, unsigned int irq_mask)
{
	unsigned int status = ifrtp_ior32(dev, offset);

	if (status & irq_mask)
		return true;

	return false;
}

bool ifrtp_get_net_raw_packet_irq_status(struct igw_priv *dev)
{
	return ifrtp_get_irq_status(dev, OFFSET_IRQ_L2_NET_STATUS, MASKBIT_IRQ_L2_NET_RAW_PACKET);
}

bool ifrtp_get_net_rx_irq_status(void)
{
	return ifrtp_get_irq_status(NULL, OFFSET_IRQ_L2_NET_STATUS, MASKBIT_IRQ_L2_NET_RX);
}

